#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import networkx as nx
import sma

def countMultiMotifs(G : nx.Graph, 
                     *arities, 
                     roles = [], 
                     iterator = None, 
                     **kwargs):
    """
    Front-end function for counting multi-motifs. The default iterator is
    :py:class:`sma.MultiMotifs`. As classificator serves :py:class:`sma.MultiMotifClassificator`.
    All options of :py:meth:`sma.countMotifs` are available.
    
    Currently supported sets of arities (signatures):
        
        - ``1, 2``: see :py:class:`sma.ThreeMotifClassificator`
        - ``2, 2``: see :py:class:`sma.FourMotifClassificator`
        - ``1, 1, 1``: see see :py:class:`sma.Multi111Classificator`
        - ``1, 1, 2``: see see :py:class:`sma.Multi121Classificator`
        - ``1, 2, 2``: see see :py:class:`sma.Multi221Classificator`
        - ``2, 2, 2``: see see :py:class:`sma.Multi222Classificator`
    
    See also :py:meth:`sma.countMultiMotifsSparse`.
    
    :param G: the graph
    :param arities: arities for the motif source and the classificator
    :param roles: roles of the level, cf. :py:class:`sma.MultiMotifClassificator`.
    :param iterator: a custom iterator, default is :py:class:`sma.MultiMotifs`.
    :param kwargs: additional parameters for :py:meth:`sma.countMotifs`
    """
    signature = sma.multiSignature(arities)
    
    # speeding up 3 and 4 motifs
    if signature == [1,2]:
        positions = sma.matchPositions([1,2], arities, roles)
        if iterator is None:
            iterator = sma.iterate3Motifs(G, positions[0], positions[1])
        return sma.count3Motifs(G, iterator, **kwargs)
    elif signature == [2,2]:
        positions = sma.matchPositions([2,2], arities, roles)
        if iterator is None:
            iterator = sma.iterate4Motifs(G, positions[0], positions[1])
        return sma.count4Motifs(G, iterator, **kwargs)
    
    # now handle all other motifs
    motif_info = sma.motifInfo(signature)
    if iterator is None:
        iterator = sma.MultiMotifsNormalized(G, *arities, roles = roles, motif_info = motif_info)
    return sma.countMotifs(G, motif_info.classificator(G), iterator, **kwargs)
        

def countMultiMotifsSparse(G : nx.Graph,
                           *arities, 
                           roles = [],
                           **kwargs):
    """
    Front-end function for several counting functions for sparse networks. It supports
    the same parameters for defining the motifs (arities, roles) as other multi-level
    counting functions. Please check the documentations of the respective functions
    as some of them might return partial counts only. The supported functions are:
        
        - :py:meth:`sma.count3MotifsSparse`
        - :py:meth:`sma.count4MotifsSparse`
        - :py:meth:`sma.count121MotifsSparse`
        - :py:meth:`sma.count221MotifsSparse`
        - :py:meth:`sma.count222MotifsSparse`
    
    See also :py:meth:`sma.countMultiMotifs`.
    
    :param G: the SEN
    :param arities: arities of the motifs, cf. :py:class:`sma.MultiMotifClassificator`.
    :param roles: roles of the levels, cf. :py:class:`sma.MultiMotifClassificator`.
    :param kwargs: additional parameters for the counter functions, see above.
    """
    signature  = sma.multiSignature(arities)
    motif_info = sma.motifInfo(signature)
    positions  = sma.matchPositions(motif_info.signature, arities, roles)
    return motif_info.sparse_counter(G, *positions, **kwargs)

def findIdealCounter(signature : list, 
                     motifs : list, 
                     assume_sparse : bool = True):
    """
    Determines an ideal counter for a family of motifs. This function checks whether
    the given motifs can be counted with countes for sparse networks 
    (if ``assume_sparse = True``) and returns in this case the counter supplemented
    with additional parameters for the counter.
    
    This function is a called by :py:meth:`sma.countMotifsAuto`.
    
    :param signature: (ordered) signature
    :param motifs: list of motif classes
    :param assume_sparse: whether the network is assumed to be sparse
    :returns: tuple consisting of
    
        - a counting function, e.g. :py:meth:`sma.countMultiMotifs` or 
          :py:meth:`sma.countMultiMotifsSparse`,
        - a dict with additional parameters for this counting function
    """
    if signature == [1,2] and assume_sparse:
        return sma.countMultiMotifsSparse, {}
    elif signature == [2,2]:
        if assume_sparse and (not('IV.A' in motifs) and 
                              not('IV.B' in motifs) and 
                              not('IV.C' in motifs) and 
                              not('IV.D' in motifs)):
            return sma.countMultiMotifsSparse, {}
    elif signature == [1,1,2]:
        if assume_sparse:
            for motif in motifs:
                if int(motif) != 3 and int(motif) != 4:
                    return sma.countMultiMotifsSparse, {}
            return sma.countMultiMotifsSparse, {'optimize_top_adjacent' : True}
    elif signature == [1,2,2]:
        if assume_sparse:
            dense = all(map(lambda m : m.endswith('.2') or m.endswith('.3'), motifs))
            if dense and all(map(lambda m : m.endswith('.3'), motifs)):
                return sma.countMultiMotifsSparse, {'optimize_top_adjacent' : True}
            return sma.countMultiMotifsSparse, {}
    elif signature == [2,2,2]:
        if assume_sparse:
            for motif in motifs:
                if int(motif) != 3 and int(motif) != 4:
                    return sma.countMultiMotifs, {}
            return sma.countMultiMotifsSparse, {}
    return sma.countMultiMotifs, {}

def countMotifsAuto(G : nx.Graph, 
                    *motifs, 
                    assume_sparse : bool = True,
                    **kwargs):
    """
    Front-end function for all counting functions in sma. This function determines
    automatically an ideal counter (sparse or dense) for a set of motifs. This 
    reduces overhead and leads to an optimization.
    
    .. code :: Python
        
        # let G by a SEN
        # count open and closed social and ecological triangles
        partial, total = sma.countMotifsAuto(G, "1,2[I.C]", "1,2[II.C]", "2,1[I.C]", "2,1[II.C]")
        print(partial)
        print(total)
    
    :param G: the SEN
    :param motifs: list of motif identifier, see documentation
    :param assume_sparse: whether the network shall be assumed to be sparse
    :param kwargs: more arguments for the counter
    :returns: tuple consisting of:
        
        - a dict with the motif identifiers as keys and the corresponding counts
          as values
        - a dict with all counts collected
    
    :raises TypeError: in case of unrecognizable motif identifiers
    """
    grouped_request = sma.groupMotifIdentifiers(*motifs)
    partial_results = {}
    total_results = {}
    
    # overwrite inadmissible parameters
    kwargs['array'] = False
    
    for cl in grouped_request.keys():
        arities, roles, _ = sma.parseMotifIdentifier(cl)
        signature = sma.multiSignature(arities)
        motif_info = sma.motifInfo(signature)
        counter, kw = sma.findIdealCounter(signature, grouped_request[cl], assume_sparse)
        counted = counter(G, *arities, roles = roles, **{**kwargs, **kw})
        total_results[cl] = counted
        # put partial result together
        for motif in grouped_request[cl]:
            # cast motif class to correct type
            key = motif_info.classes_type(motif)
            if key not in counted:
                raise TypeError("Unrecognizable motif class in identifier %s[%s]."\
                                % (cl, motif))
            partial_results["%s[%s]" % (cl, motif)] = counted[key]
    
    return partial_results, total_results

def exemplifyMotif(G : nx.Graph, identifier : str) -> tuple:
    """
    Returns an example for a motif with given identifier in a SEN.
    
    :param G: the SEN
    :param identifer: motif identifier string
    :returns: motif tuple in normalized form (ordered as in signature)
    :raises StopIteration: if the SEN does not contain such a motif
    """
    arities, roles, motif = sma.parseMotifIdentifier(identifier)
    signature = sma.multiSignature(arities)
    motif_info = sma.motifInfo(signature)
    source = sma.MultiMotifsNormalized(G, *arities, roles = roles, motif_info = motif_info)
    motif = motif_info.classes_type(motif)
    selector = sma.isClass(motif_info.classificator(G), motif)
    return next(source & selector)


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
import sma
import inspect, sys
import itertools

class MotifInfo:
    """
    Parent class of all MotifInfo classes. Subclasses contain information about
    motifs, their classificators and counters. The following properties are provided:
        
    :var signature: signature of the motifs
    :var classes: list of names of all motif classes
    :var classes_type: type of the class names, e.g. ``str`` or ``int``
    :var classificator: subclass of :py:class:`sma.MotifClassificator` for 
          classifying motifs
    :var counter: counting function with signature
        
          ``counter(G, *levels, **kwargs)``
    :var sparse_counter: counting function for sparse graphs with signature
        
          ``sparse_counter(G, *levels, **kwargs)``
    :var expectations: function for computing motif expectations in a
          Erdős-Rényi model
        
          ``expectations(G, *levels, **kwargs)``
    :var variances: function for computing motif variances in a Erdős-Rényi model
        
          ``variances(G, *levels, **kwargs)``
    :var projections: dict containing information about closed/open pairs of motifs,
        i.e. which motif classes belong together when non considering dyads on a 
        specific level of the motif. The dict is indexed by abstract levels (indexes
        of levels in ``signature``). Every entry is a list of tuples of motif class
        names. Each tuple contains first the open and secondly the closed motif.
        See :py:class:`sma.Motif3Info` as an example.
    
    If a certain feature is not implemented, the corresponding value is set to ``None``.
    """
    signature      = []
    classes        = []
    classes_type   = None
    classificator  = None
    counter        = None
    sparse_counter = None
    expectations   = None
    variances      = None
    projections    = {}

class Motif3Info(MotifInfo):
    signature      = [1, 2]
    classes        = sma.MOTIF3_NAMES
    classes_type   = str
    classificator  = sma.ThreeMotifClassificator
    counter        = sma.count3Motifs
    sparse_counter = sma.count3MotifsSparse
    expectations   = sma.expected3Motifs
    variances      = sma.var3Motifs
    projections    = {1: [('I.A', 'II.A'),
                          ('I.B', 'II.B'),
                          ('I.C', 'II.C')]}

class Motif4Info(MotifInfo):
    signature      = [2, 2]
    classes        = sma.MOTIF4_NAMES
    classes_type   = str
    classificator  = sma.FourMotifClassificator
    counter        = sma.count4Motifs
    sparse_counter = sma.count4MotifsSparse
    expectations   = sma.expected4Motifs
    variances      = None
    projections    = {0: list(zip(sma.MOTIF4_NAMES[::2], sma.MOTIF4_NAMES[1::2])),
                      1: list(zip(sma.MOTIF4_NAMES[::4] + sma.MOTIF4_NAMES[1::4], 
                                  sma.MOTIF4_NAMES[2::4] + sma.MOTIF4_NAMES[3::4]))}

class Motif111Info(MotifInfo):
    signature      = [1, 1, 1]
    classes        = [0, 1, 2, 3, 4, 5, 6, 7]
    classes_type   = int
    classificator  = sma.Multi111MotifClassificator
    counter        = None
    sparse_counter = None
    expectations   = None
    variances      = None

class Motif121Info(MotifInfo):
    signature      = [1, 2, 1]
    classes        = [-1, 1, 2, 3, 4]
    classes_type   = int
    classificator  = sma.Multi121MotifClassificator
    counter        = None
    sparse_counter = sma.count121MotifsSparse
    expectations   = sma.expected121Motifs
    variances      = None

class Motif221Info(MotifInfo):
    signature      = [2, 2, 1]
    classes        = ['Unclassified'] + list(map(lambda x : '%s.%d' % x, 
                     itertools.product(sma.MOTIF4_NAMES, [0,1,2,3])))
    classes_type   = str
    classificator  = sma.Multi221MotifClassificator
    counter        = None
    sparse_counter = sma.count221MotifsSparse
    expectations   = sma.expected221Motifs
    variances      = None

class Motif222Info(MotifInfo):
    signature      = [2, 2, 2]
    classes        = [-1, 1, 2, 3, 4]
    classes_type   = int
    classificator  = sma.Multi222MotifClassificator
    counter        = None
    sparse_counter = sma.count222MotifsSparse
    expectations   = sma.expected222Motifs
    variances      = None

class Motif2Info(MotifInfo):
    signature      = [2]
    classes        = [0, 1]
    classes_type   = int
    classificator  = sma.TwoMotifClassificator
    counter        = None
    sparse_counter = None
    expectations   = None
    variances      = None

class Motif11Info(MotifInfo):
    signature      = [1, 1]
    classes        = [0, 1]
    classes_type   = int
    classificator  = sma.TwoMotifClassificator
    counter        = None
    sparse_counter = None
    expectations   = None
    variances      = None
    
class Motif3pInfo(MotifInfo):
    signature      = [3]
    classes        = [0, 1, 2, 3]
    classes_type   = int
    classificator  = sma.ThreePMotifClassificator
    counter        = None
    sparse_counter = None
    expectations   = None
    variances      = None

def _infoObjects():
    return filter(lambda info: issubclass(info[1], MotifInfo) and info[1] != MotifInfo,
                  inspect.getmembers(sys.modules[__name__], inspect.isclass))
    
def motifInfo(signature):
    """
    Returns a subclass of :py:class:`sma.MotifInfo` correponding to the given
    signature. The order of the signature is irrelevant.
    
    :param signature: the signature
    :returns: subclass of :py:class:`sma.MotifInfo`
    :raises NotImplementedError: if no subclass of :py:class:`sma.MotifInfo` is
        available for the given signature
    """
    signature = sorted(signature)
    for name, obj in _infoObjects():
        if signature == sorted(obj.signature):
            return obj
    raise NotImplementedError('motifs with signature %s not implemented' % str(signature))

def supportedSignatures() -> iter:
    """
    Returns an iterator yielding all supported signatures.
    
    .. code :: Python
        
        print(list(sma.supportedSignatures()))
        
    
    :returns: iterator yielding all supported signatures.
    """
    return map(lambda x : x[1].signature, _infoObjects())



"""
Social-Ecological-System Motif Analyser (SMA).

The module provides the following main functions:
    - reading SES from CSV files
    - analysing SES', especially counting 3- and 4-motifs with or without
      taking nodal attributes into account
    - generating random SES'
    - exchanging information with R (using statnet on the R side)

"""

# load core module first
from .properties import *
from .helper import *
from .io import *
from .classify import *
from .iterate import *
from .analyse import *
from .generate import *

# load high level modules
from .analyse_frontend import *
from .analyse_triangles import *
from .analyse_cooccurrence import *
from .analyse_distribution import *
from .analyse_simulate import *
from .draw import *
from .rbridge import *

# create info database
from .motif_infos import *


# -*- coding: utf-8 -*-

import functools
import networkx as nx
import numpy

def binaryArrayHornersMethod(array : list) -> int:
    """
    transforms the binary number encoded in the given array
    to an integer
    
    :example: binaryArrayHornersMethod([1,0,1]) = 5
    """
    return functools.reduce(lambda x, y: x*2 +y, array)

def multiToWeightedGraph(M : nx.MultiGraph, attr = 'weight') -> nx.Graph:
    """
    Helper function for converting a multigraph into a weighted simple graph with
    edge weights corresponding to the multiplicities of the edges in the multigraph.
    
    Nodal attributes are preserved while edge attributes will be lost. Use the 
    second parameter to set the name of the weight attribute.
    
    :param M: a multigraph
    :param attr: name of the weight attribute
    :returns: a simple graph with weighted edges
    """
    R = nx.Graph()
    R.add_nodes_from(M.nodes(data=True))
    for u,v,data in M.edges(data=True):
        if R.has_edge(u,v):
            R[u][v][attr] += 1
        else:
            R.add_edge(u, v, **{attr : 1})
    return R

def maxEdgeCountMatrix(nVertices : numpy.ndarray) -> numpy.ndarray:
    """
    Returns a matrix containing the maximal possible number of edges in the
    subsystems of a multigraph. The entries of this matrix constitute the upper 
    bounds of the entries of the ``nEdges`` parameter of :py:meth:`sma.randomMultiSENs`.
    
    Let :math:`v_1, \dots, v_n` be the entries of ``nVertices``. Then the entries
    of the returned matrix are
    
    .. math:: 
        
        a_{ij} = \\begin{cases} \\frac12 v_i (v_i-1), & i = j \\\\ v_i v_j, & i < j \\end{cases}.
    
    See also :py:meth:`sma.motifClassMatrix`.
    
    :param nVertices: numbers of vertices per subsystem
    """
    N = len(nVertices)
    matrix = numpy.diag([n*(n-1)//2 for n in nVertices])
    for i in range(N):
        for j in range(i+1, N):
            matrix[i][j] = nVertices[i]*nVertices[j]
    return matrix

def randomEdgeCountMatrix(nVertices : numpy.ndarray) -> numpy.ndarray:
    """
    This function takes the result of :py:meth:`sma.maxEdgeCountMatrix` and a matrix
    with random integer entries less or equal this matrix. It can be used to generate
    a random multilevel SEN with :py:meth:`sma.randomMultiSENs`.
    
    :param nVertices: numbers of vertices per subsystem
    """
    matrix = maxEdgeCountMatrix(nVertices)
    random = numpy.random.random_sample(numpy.shape(matrix))
    return numpy.array(numpy.multiply(matrix, random), dtype=int)

def splitMotifIdentifer(identifier : str) -> tuple:
    """
    Splits a motif identifier into the first part (encoding the pattern and roles 
    of the motif) and the second part (encoding the type of the motif).
    
    For example, ``1:2,2:1[I.C]`` is mapped to the tuple consisting of ``1:2,2:1``
    and ``I.C``.
    
    :param identifier: motif identifier string
    :returns: tuple consisting of first and second part
    """
    if '[' in identifier:
        first = identifier[:identifier.index('[')]
        second = identifier[identifier.index('[')+1:identifier.index(']')]
    else:
        first = identifier
        second = None
    return (first, second)
        
def parseMotifIdentifier(identifier : str) -> tuple:
    """
    Parser for motif identifier strings. This function extracts a patterns, roles 
    and motif from a motif identifier string, cf. :py:class:`sma.MultiMotifClassificator`.
    
    A simple motif identifier, e.g. ``1,2[I.C]``, consist of a comma-separated 
    list of integers specifying the pattern of the motif (here, 1 node from one level,
    2 nodes from the other), and of an optional motif name enclosed in brackets.
    
    Motif identifiers can also be used to specify the roles of the levels. An example
    is ``1:1,2:0[I.C]``. The values behind the colons are the roles of the levels.
    
    :param identifier: motif identifier string
    :returns: triple consisting of:
        
        - an arities array
        - a roles array, or ``[]`` if no roles are specified
        - a motif, or ``None`` if no motif is specified
    
    :raises TypeError: in case of formatting problems
    :raises AssertionError: in case of formatting problems
    """
    first, second = splitMotifIdentifer(identifier)
    if ':' in first:
        pairs = first.split(",")
        arities = []
        roles = []
        for p in pairs:
            x, y = p.split(":")
            assert int(x) >= 1, 'illegal motif identifier: value %d as signature entry not allowed' % int(x)
            arities.append(int(x))
            roles.append(int(y))
            assert int(y) >= 0, 'illegal motif identifier: value %d as role not allowed' % int(y)
        return (arities, roles, second)
    else:
        vals = first.split(",")
        return (list(map(int, vals)), [], second)

def groupMotifIdentifiers(*motifs : str) -> dict:
    """
    Groups motif identifiers by their heads in a dictionary. For example,
    the list ``1,2[I.A], 2,2[II.B], 1,2[II.B]`` is mapped to 
    ``{'1,2': ['I.A', 'II.B'], '2,2': ['II.B']}``. The returned dict is indexed
    by the heads of the identifers, the values are lists of motif classes.
    
    :param motifs: list of motif identifiers
    :returns: dict of grouped motif identifiers
    """
    grouped = {}
    for motif in motifs:
        f, s = splitMotifIdentifer(motif)
        if f in grouped:
            grouped[f].append(s)
        else:
            grouped[f] = [s]
    return grouped

def multiSignature(arities : list) -> list:
    """
    Returns the signature of a list of arities. The signature is the list
    of non-zero entries in ascending order. It determines for example the counter 
    function that can be used to count motifs with these arities.
    
    :param arities: list of arities
    :returns: signature
    """
    return sorted(filter(lambda a : a != 0, arities))

def matchPositions(signature : list, 
                   arities : list, 
                   roles : list = []) -> list:
    """
    Position matching is process in which a signature (``1,2``-motifs) is mapped
    to list of positions, i.e. ``sesTypes`` in the network (:py:const:`sma.NODE_TYPE_ECO`,
    :py:const:`sma.NODE_TYPE_SOC` in case of ecological triangles). Optionally,
    a list of roles can be provides. This overwrites the position matching.
    
    :param pattern: the signature
    :param arities: the arities given
    :param roles: optional list of rolls
    :returns: positions
    :raises AssertionError: in case of mismatching signature/arities/roles
    """
    if len(roles) != 0:
        # with roles
        assert len(signature) == len(roles), \
            'signature %s and roles %s must be of the same length' % (str(signature), str(roles))
        if signature == arities:
            return roles
        else:
            arities_positive = list(filter(lambda a : a > 0, arities))
            assert len(arities_positive) == len(signature), \
                'signature %s and positive arities %s must be of the same length' %  (str(signature), str(arities_positive))
            prepositions = matchPositions(signature, arities_positive)
            return [roles[i] for i in prepositions]
    else:
        # without roles
        positions = []
        last = {}
        for p in signature:
            if p in last:
                last[p] = arities.index(p, last[p]+1)
            else:
                last[p] = arities.index(p)
            positions.append(last[p])
        return positions



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import itertools, functools
import scipy.special
import numpy
import networkx as nx
import sma

def triangleCoefficient(G : nx.Graph, node, othertype = None) -> float:
    """
    Calculates the triangle coefficient of node, which is defined as follows:
    
    Take all 3-motifs with the node as distinct node and compute the number of 
    I.C and II.C motifs (open and closed triangles with the given node at the point).
    Divide the number of I.C motifs by the sum of number of I.C motifs and II.C motifs.
    
    See :py:meth:`sma.cooccurrenceTable` for a more general function.
    
    :param G: the graph
    :param node: the node whose triangle coefficient shall be computed
    :param othertype: the ``sesType`` of the non-distinct nodes or ``None`` if all
        nodes with ``sesType`` different from the ``sesType`` of the distinct node
        shall be considered. A value must be specified when used with multi-level
        networks in order to avoid unexpected behaviour.
    :returns: II.C / (I.C + II.C), i.e. the ratio of closed triangles vs. all triangles,
        if no triangles are found, this function returns zero.
    """
    neighbors = G[node]
    nodesType = G.nodes[node]['sesType']
    check = lambda n : G.nodes[n]['sesType'] != nodesType
    if othertype is not None:
        check = lambda n : G.nodes[n]['sesType'] == othertype
    neighborsOtherType = list(filter(check, neighbors))
    countTriangles = scipy.special.binom(len(neighborsOtherType), 2)
    if countTriangles == 0:
        return 0
    countClosedTriangles = sum(map(lambda pair: G.has_edge(*pair), itertools.combinations(neighborsOtherType, 2)))
    return countClosedTriangles / countTriangles

def triangleCoefficients(G : nx.Graph, array : bool = False):
    """
    Computes the triangle coefficients for all nodes in a graph.
    
    :param G: the graph
    :param array: set array to True if you want the result as a list instead of a dict
    
    See :py:meth:`sma.triangleCoefficient` for details.
    """
    if array:
        return numpy.fromiter((triangleCoefficient(G, node) for node in G.nodes), dtype = float)
    else:
        return {node : triangleCoefficient(G, node) for node in G.nodes}

def connectedOpposingSystem(G : nx.Graph, node, othertype = None) -> set:
    """
    Returns the set of nodes which are connected to the given node but have the
    opposite social-ecological type.
    
    :param G: the graph
    :param node: the node
    :param othertype: the ``sesType`` of the opposite nodes or ``None`` if all
        nodes with ``sesType`` different from the ``sesType`` of the specified node
        shall be considered. A value must be specified when used with multi-level
        networks in order to avoid unexpected behaviour.
    :returns: set of nodes which share edges with the given node and are of the 
        other social-ecological type.
    """
    neighbors = G[node]
    nodesType = G.nodes[node]['sesType']
    check = lambda n : G.nodes[n]['sesType'] != nodesType
    if othertype is not None:
        check = lambda n : G.nodes[n]['sesType'] == othertype
    neighborsOtherType = set(filter(check, neighbors))
    return neighborsOtherType

def overlappingCoefficients(G : nx.Graph, subsystem : int = sma.NODE_TYPE_ECO) -> numpy.ndarray:
    """
    Returns a matrix of overlapping coefficients for a given subsystem of nodes.
    For two nodes :math:`x`, :math:`y` of the same type the overlapping coefficient is defined as
    
    .. math ::
        
        O(x,y) = \\begin{cases}
                    \\frac{|S(x) \cap S(y)|}{|S(x) \cup S(y)|}, & \\text{if } S(x) \cup S(y) \\neq \\emptyset \\\\
                    0, & \\text{otherwise} \\
                 \\end{cases}
        
    where :math:`O(x,y)` is the overlapping coefficient of :math:`x` and :math:`y` and :math:`S(i)` 
    is the connected opposing system of :math:`i`, cf. :py:meth:`sma.connectedOpposingSystem`, i.e. the 
    set of nodes which are connected to :math:`i` but are of the opposite social ecological
    type.
    
    :param G: the graph
    :param subsystem: type of the nodes to compute the coefficients for, default is
        :py:const:`sma.NODE_TYPE_ECO`.
    :returns: symmetric matrix containing the coefficients, default is
        :py:const:`sma.NODE_TYPE_ECO` for computing the overlapping coefficients
        of the ecological subsystem
    
    """
    nodes = list(sma.sesSubgraph(G, subsystem))
    result = numpy.identity(len(nodes))
    opposingSystems = list(map(functools.partial(connectedOpposingSystem, G), nodes))
    
    for i in range(len(nodes)):
        A = opposingSystems[i]
        if len(A) == 0:
            continue
        for j in range(i):
            if G.has_edge(nodes[i], nodes[j]):
                B = opposingSystems[j]
                result[i,j] = result[j, i] = len(A.intersection(B))/ len(A.union(B))
        
    return result


#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import networkx as nx
import sma
import pandas

def translateGraph(adjacency, 
                   attributeNames, 
                   attributeValues, 
                   typeAttr : str = None) -> nx.Graph:
    """
    Returns a :py:class:`networkx.Graph` from a adjacency matrix, a list of nodal attribute names
    and a matrix of nodal attribute values. This function is designed to make networkx objects
    compatible with R's statnet network objects.
    
    :param adjacency: adjacency matrix provided as a numerical matrix
    :param attributeNames: list of attribute names
    :param attributeValues: matrix of attribute values in accordance to `attributeNames`.
    :param typeAttr: name of an attribute with specifies whether a node is social or ecological.
        The values must be integers, matching :py:const:`sma.NODE_TYPE_SOC` and :py:const:`sma.NODE_TYPE_ECO` in
        two-level networks. If None, attributeNames must contain 'sesType'.
    :returns: networkx graph with the given properties
    """
    G = nx.from_numpy_matrix(adjacency, create_using=nx.Graph())
    
    assert ((typeAttr is not None and typeAttr in attributeNames) or 
           (typeAttr is None and 'sesType' in attributeNames)), 'sesType attribute must be provided'
    
    for attr, values in zip(attributeNames,attributeValues):
        nx.set_node_attributes(G, name=attr, values=dict(enumerate(values)))
        if typeAttr == attr:
            nx.set_node_attributes(G, name='sesType', values={i : int(v) for (i,v) in enumerate(values)})
    
    return G

def countAnyMotifs(it : sma.ComplexMotifIterator) -> int:
    """
    Counts the number of elements in the given iterator.
    
    Function for compatability with R.
    """
    try:
        return len(it)
    except TypeError:
        return sum(1 for _ in it)

def motifSet(source : sma.SourceMotifIterator, 
             *conditions : sma.ConditionMotifIterator) -> sma.ComplexMotifIterator:
    """
    Returns an motif interator representing a set of motifs with 
    certain properties.
    
    Function for compatability with R.
    
    :param source: :py:class:`sma.SourceMotifIterator`, the basal set to take the
        motifs from
    :param conditions: one or several :py:class:`sma.ConditionMotifIterator` 
        representing conditions which the motifs must fulfill to be part of the set
    """
    it = source
    for cond in conditions:
        it = it & cond
    return it

def listMotifs(iterator : sma.ComplexMotifIterator) -> list:
    """
    Returns a list of all motifs in a motif iterator.
    
    Function for compatability with R.
    
    :param iterator: motif iterator
    """
    return list(iterator)

def _gatherMotifDict(d : dict) -> dict:
    result = {}
    for head, d2 in d.items():
        for motif, value in d2.items():
            result['%s[%s]' % (head, str(motif))] = value
    return result

def _ensureInt(args : dict, keys : list) -> dict:
    for key in keys:
        if key in args:
            args[key] = int(args[key])
    return args

def countMotifsAutoR(G : nx.Graph, 
                     motifs : list, 
                     omit_total_result : bool = False,
                     **kwargs) -> pandas.DataFrame:
    """
    Wrapper function for :py:meth:`sma.countMotifsAuto`. Optimized for the reticulate
    R interface.
    
    :param G: the SEN
    :param motifs: the list of motifs to be counted
    :param omit_total_result: whether only the partial result shall be returned
    :param kwargs: further parameters for :py:meth:`sma.countMotifsAuto`
    :returns: pandas dataframe with either partial or total counts
    """
    partial, total = sma.countMotifsAuto(G, *motifs, **kwargs)
    if omit_total_result:
        df = pandas.DataFrame(data = partial, index = ['count'])
        return df
    else:
        gathered = _gatherMotifDict(total)
        df = pandas.DataFrame(data = gathered, index = ['count'])
        return df

def distributionMotifsAutoR(G : nx.Graph, 
                            motifs : list,
                            model : str = sma.MODEL_ERDOS_RENYI,
                            level : int = -1,
                            omit_total_result : bool = False) -> pandas.DataFrame:
    """
    Wrapper function for :py:meth:`sma.distributionMotifsAuto`. Optimized for the reticulate
    R interface.
    
    :param G: the SEN
    :param motifs: the list of motifs whose distributions are to be described
    :param omit_total_result: whether only the partial result shall be returned
    :returns: pandas dataframe with either partial or total distribution
        information
    """
    level = int(level)
    partial, total = sma.distributionMotifsAuto(G, *motifs, model = model, level = level)
    if omit_total_result:
        df = pandas.DataFrame(data = partial, index = ['expectation', 'variance'])
        return df
    else:
        gathered = _gatherMotifDict(total)
        df = pandas.DataFrame(data = gathered, index = ['expectation', 'variance'])
        return df

def simulateBaselineAutoR(G : nx.Graph,
                          motifs : list,
                          **kwargs) -> pandas.DataFrame:
    """
    Wrapper function for :py:meth:`sma.simulateBaselineAutoR`.
    Optimized for the reticulate R interface.
    
    :param G: the SEN
    :param motifs: the list of motifs whose distributions are to be described
    :param kwargs: further parameters for :py:meth:`sma.simulateBaselineAutoR`.
    :returns: pandas dataframe with one column for every motif and one row for 
        every random SEN
    """
    kwargs = _ensureInt(kwargs, ['n'])
    simulation = sma.simulateBaselineAuto(G,
                                          *motifs,
                                          **kwargs)
    df = pandas.DataFrame(data = simulation)
    return df

def identifyGapsR(G : nx.Graph, 
                  motif_identifier : str, 
                  level : int = -1) -> pandas.DataFrame:
    """
    Wrapper function for :py:meth:`sma.indentifyGapsR`. Optimized for the reticulate
    R interface.
    
    :param G: the SEN
    :param motif_identifier: a motif identifier
    :param level: level
    :returns: pandas dataframe with three columns: two for the edges, one for their
        contribution
    """
    level = int(level)
    result = sma.identifyGaps(G, motif_identifier, level = level)
    df = pandas.DataFrame(map(lambda x: (*x[0],x[1]), result), 
                          columns=['vertex0', 'vertex1', 'contribution'])
    return df

def isClosedMotifR(motif_identifier : str, level : int = -1) -> bool:
    """
    Wrapper function for :py:meth:`sma.isClosedMotif`. Optimized for the reticulate
    R interface.
    
    :param motif_identifier: motif identifier
    :param level: level
    :returns: whether the motif is closed with respect to the specified level
    :raises NotImplementedError: if open/closed relations are not implemented for
        this motif and level
    """
    return sma.isClosedMotif(motif_identifier, int(level))


# -*- coding: utf-8 -*-

import itertools
import multiprocessing
import numpy
import networkx as nx
import scipy.special

import sma

# Counters
def countMotifs(G : nx.Graph, 
                classificator : sma.MotifClassificator, 
                iterator : sma.MotifIterator, 
                processes : int = 0,
                norm : bool = False,
                array : bool = False,
                chunksize : int = 10000):
    """
    General function for classifying and counting motifs. 
    
    Procedure:
        1. motifs are drawn from the iterator
        2. classificator is called for every motif
        3. the number of occurences is computed for every class defined in classes
    
    Note, that this function provides for must cases to many options. See
    :py:meth:`sma.count4Motifs`, :py:meth:`sma.count3EMotifs` and :py:meth:`sma.count3SMotifs`.
    
    This function supports multiprocessing. Set processes to any value other than zero
    to use several processes to solve the task. Set processes to None to use 
    :py:class:`multiprocessing.pool.Pool`'s default value.
    
    Set norm to True to get normalized counts, i.e. in this case the sum of all counts
    equals one.
    
    Set array to True to get an array where the indices match with the indices of the classes
    in classes instead of a dict mapping classes to their respective counts.
    
    :param G: the graph
    :param classificator: a :py:class:`sma.MotifClassificator` for classifying the
        motifs
    :param iterator: an iterator yielding motifs. Note, that the format of the motifs must be
        somehow compatible with the format used by the classificator. See :py:func:`sma.iterate4Motifs` etc.
    :param processes: number of processes the task is split among for multiprocessing. Set processes to zero
        to disable multiprocessing.
    :param chunksize: in multiprocessing mode size of each chunk that is sent to
        a child
    :param norm: whether the counts should be normalized in the way that they sum up to one.
    :param array: whether the results should be returned as array (indices map the indices in classes) or
        as dict mapping a class to its count
    :returns: number of occurences of the specified classes in either an array or a dict, cf. parameter array

    """
    classes = classificator.info().classes
    result = {key : 0 for key in classes}
    
    if processes == 0:
        classified = map(classificator, iterator)
        for cl in classified:
            result[cl] += 1
            del cl
    else:
        with multiprocessing.Pool(processes) as p:
            classified = p.imap_unordered(classificator, iterator, chunksize = chunksize)
            for cl in classified:
                result[cl] += 1
                del cl
            p.close()
            p.join()
    
    # normalize
    if norm:
        factor = 1. / sum(result.values())
        result = {key : factor * result[key] for key in result}

    # output format    
    if array:
        return numpy.array([result[k] for k in classes])
    
    return result

def count4Motifs(G : nx.Graph, iterator = None, **kwargs):
    """
    Front-end function for counting 4-motifs in a given graph.
    If iterator is None (default) :py:meth:`sma.iterate4Motifs` is used.
    
    See :py:meth:`sma.countMotifs` for details. See also :py:meth:`sma.count4MotifsSparse`.
    
    :param G: the graph
    :param iterator: a custom iterator, if None (default) :py:meth:`sma.iterate4Motifs` is used.
    :param kwargs: additional parameters for :py:meth:`sma.countMotifs`
    """
    if iterator is None:
        iterator = iterate4Motifs(G)
    return countMotifs(G, sma.FourMotifClassificator(G), iterator, **kwargs)

def count3Motifs(G : nx.Graph, iterator, **kwargs):
    """
    Front-end function for counting 3-motifs in a given graph.
    
    See :py:meth:`sma.countMotifs` for details. See also :py:meth:`sma.count3MotifsSparse`.
    
    :param G: the graph
    :param iterator: an iterator which provides 3-motifs
    :param kwargs: additional parameters for :py:meth:`sma.countMotifs`
    """
    return countMotifs(G, sma.ThreeMotifClassificator(G), iterator, **kwargs)

def count3EMotifs(G : nx.Graph, iterator = None, **kwargs):
    """
    Front-end function for counting 3E-motifs (distinct ecological node) in a given graph.
    If iterator is None (default) :py:meth:`sma.iterate3EMotifs` is used.
    
    See :py:meth:`sma.countMotifs` for details. See also :py:meth:`sma.count3MotifsSparse`.
    
    :param G: the graph
    :param iterator: a custom iterator, if None (default) :py:meth:`sma.iterate3EMotifs` is used.
    :param kwargs: additional parameters for :py:meth:`sma.countMotifs`
    
    """
    if iterator is None:
        iterator = iterate3EMotifs(G)
    return count3Motifs(G, iterator, **kwargs)

def count3SMotifs(G : nx.Graph, iterator = None, **kwargs):
    """
    Front-end function for counting 3S-motifs (distinct social node) in a given graph.
    If iterator is None (default) :py:meth:`sma.iterate3SMotifs` is used.
    
    See :py:meth:`sma.countMotifs` for details. See also :py:meth:`sma.count3MotifsSparse`.
    
    :param G: the graph
    :param iterator: a custom iterator, if None (default) :py:meth:`sma.iterate3SMotifs` is used.
    :param kwargs: additional parameters for :py:meth:`sma.countMotifs`
    """    
    if iterator is None:
        iterator = iterate3SMotifs(G)
    return count3Motifs(G, iterator, **kwargs)

def count3MotifsSparse(G : nx.Graph, 
                       level0, 
                       level1, 
                       array : bool = False):
    """
    Alternative function for counting 3-motifs optimized for sparse SENs, i.e.
    networks with :math:`|E| \ll |V|^2` for :math:`|E|` number of edges, :math:`V`
    number of vertices.
    
    This function counts 3-motifs with an asymptotic execution time of :math:`O(|E||V|)`.
    Multilevel networks are supported. Use parameters ``level0`` (resp. ``level1``)
    to specify the ``sesType`` of the distinct node (resp. the other nodes).
    
    The performance improvement is achieved by iterating over all cross-level edges 
    and in an inner loop over all vertices in ``level1`` distinct from the node 
    contained in the edge. Then the following arithmetic steps are performed:
        
        - the number of I.C and II.C motifs is halved since they are counted twice
          due to their symmetry
        - the total number of edges that have been already encountered in the loop
          described above is computed. Here :py:const:`sma.MOTIF3_EDGES` is used.
        - it remains only to adjust the number of I.A (no edges at all) and of I.C
          (one edge in ``level1`` only) motifs. The number of I.C motifs is the difference
          of the total number of ``level1``-``level1`` edges counted with multiplicity
          and the number of already encountered edges of this type
        - the number of I.A motifs is the total number of 3-motifs, cf. :py:meth:`totalMultiMotifs`,
          and the amount of motifs already encountered
    
    The penultimate step is based on the following observation: There are six different
    classes of 3-motifs. Let :math:`m \in \mathbb{N}^6` be a row vector denoting the numbers of
    occurences of the motif classes. Let :math:`C \in \mathbb{N}^{6 \\times 2}` denote 
    the matrix in :py:const:`sma.MOTIF3_EDGES` containing one row for every motif
    class with two entries each. The first entry represents the number of ``level1``
    edged contained in this motif, the second the number of ``level0`` edges. Let
    :math:`V_i` denote the set of vertices in level :math:`i`, :math:`E_{i,j}` the 
    set of edges from level :math:`i` to level :math:`j`. Then we have the following
    equalities
    
    .. math ::
        
        (m C)_{0} = |V_0| |E_{1,1}|, \quad 
        (m C)_{1} = (|V_1| - 1) |E_{0,1}|.
    
    Both follow using a double counting principle.
    
    Warning: The input graph must not contain self loops. Check using 
    :py:meth:`networkx.nodes_with_selfloops`.
    
    See also :py:meth:`sma.count3Motifs`.
    
    :param G: the SEN
    :param level0: ``sesType`` of the distinct node
    :param level1: ``sesType`` of the other nodes
    :param array: whether the output should be an array or a dict
    :returns: array or dict with numbers of 3-motifs
    """
    result = {key : 0 for key in sma.MOTIF3_NAMES}
    
    H = G.subgraph(filter(lambda x : G.nodes[x]['sesType'] == level0 or G.nodes[x]['sesType'] == level1, G.nodes))
    data = H.nodes(data=True)
    
    edges_count11 = 0
    edges_count01 = 0
    
    for edge in H.edges():
        v1, v2 = edge
        if data[v1]['sesType'] != data[v2]['sesType']:
            edges_count01 += 1
            e1 = v1 if data[v1]['sesType'] == level1 else v2
            s  = v1 if data[v1]['sesType'] == level0 else v2
            for e2 in sma.sesSubgraph(H, level1):
                if e1 != e2:
                    result[sma.classify3Motif(G, (s, e1, e2))] += 1
        elif data[v1]['sesType'] == level1 and data[v2]['sesType'] == level1:
             edges_count11 += 1
    # remove symmetry
    result['I.C'] //= 2
    result['II.C'] //= 2

    result_array = numpy.array([result[key] for key in sma.MOTIF3_NAMES])
    counts = result_array @ sma.MOTIF3_EDGES
    V = sma.nodesCount(H)
    
    result_array[3] = V[level0] * edges_count11 - counts[0]
    result_array[0] = V[level0] * V[level1] * (V[level1] - 1) // 2 - sum(result_array)
    
    if not array:
        return {k : v for k, v in zip(sma.MOTIF3_NAMES, result_array)}
    return result_array

def count3SMotifsSparse(G : nx.Graph, **kwargs):
    """
    Wrapper for :py:meth:`sma.count3MotifsSparse`.
    :param G: the SEN
    :param kwargs: optional arguments for :py:meth:`sma.count3MotifsSparse`.
    """
    return count3MotifsSparse(G, sma.NODE_TYPE_SOC, sma.NODE_TYPE_ECO, **kwargs)
def count3EMotifsSparse(G : nx.Graph, **kwargs):
    """
    Wrapper for :py:meth:`sma.count3MotifsSparse`.
    :param G: the SEN
    :param kwargs: optional arguments for :py:meth:`sma.count3MotifsSparse`.
    """
    return count3MotifsSparse(G, sma.NODE_TYPE_ECO, sma.NODE_TYPE_SOC, **kwargs)

def count4MotifsSparse(G : nx.Graph, 
                       level0 = sma.NODE_TYPE_SOC, 
                       level1 = sma.NODE_TYPE_ECO, 
                       guess_IV_A : int = -1, 
                       array : bool = False):
    """
    Alternative function for counting 4-motifs optimized for sparse SENs, i.e.
    networks with :math:`|E| \ll |V|^2` for :math:`|E|` number of edges, :math:`V`
    number of vertices.
    
    This function counts 4-motifs with an asymptotic execution time of :math:`O(|E||V|^2)`.
    Multilevel networks are supported. Use parameters ``level0`` (resp. ``level1``)
    to specify the ``sesType`` of the upper level (resp. the lower level) or keep
    the default values to count classical 4-motifs, i.e. as counted by 
    :py:meth:`sma.count4Motifs`.
    
    See :py:meth:`sma.count3MotifsSparse` for a more detailed description of the
    procedure.
    
    **Computing the numbers of motifs IV.A-IV.D:** This function iterates over
    cross-level edges (and in an inner loop over one node from each of the two levels). 
    This means that motifs without cross-level edges (IV.A to IV.D) cannot be recognized.
    However the number of these motifs can be reconstructed based on the numbers
    for the other motifs. After completing the iteration we know the numbers of 
    all motifs except for the classes IV.A to IV.D. Based on the double counting 
    formulas outlined in :py:meth:`sma.count3MotifsSparse` we can compute the number
    of edges :math:`E_i` in level :math:`i` that we have not encountered yet. 
    The number of remaining motifs :math:`M` can be easily computed as well 
    following :py:meth:`sma.total4Motifs`. Then the following linear equations hold.

    .. math ::
        
        \\text{IV.B} + \\text{IV.D} &= E_0  \\\\
        \\text{IV.C} + \\text{IV.D} &= E_1  \\\\
        \\text{IV.A} + \\text{IV.B} + \\text{IV.C} + \\text{IV.D} &= M \\ 
    
    Unfortunately, this system of linear equation does not admit a unique solution.
    However, if the number of IV.A motifs is known (specified using parameter
    ``guess_IV_A``), then the system can be easily solved by setting
    :math:`M' = M - \\mathtt{guess\\_IV\\_A}` and computing

    .. math ::
        
        \\left(\\begin{matrix}\\text{IV.B}\\\\\\text{IV.C}\\\\\\text{IV.D}\\end{matrix}\\right)
        =
        \\left(\\begin{matrix}1&0&1\\\\0&1&1\\\\1&1&1\\end{matrix}\\right)^{-1}
        \\left(\\begin{matrix}E_0\\\\E_1\\\\M'\\end{matrix}\\right)
        =
        \\left(\\begin{matrix}0&-1&1\\\\-1&0&1\\\\1&1&-1\\end{matrix}\\right)
        \\left(\\begin{matrix}E_0\\\\E_1\\\\M'\\end{matrix}\\right).
        
    If ``guess_IV_A`` is negative, the values for IV.A-IV.D will be set to :math:`-1`.
    Note that in the case the sum of all returned values does not equal the integer
    returned by :py:meth:`sma.total4Motifs` any more.
    
    See also :py:meth:`sma.count4Motifs`.
    
    :param G: the SEN
    :param level0: ``sesType`` of the lower nodes, social in classical 4-motifs
    :param level1: ``sesType`` of the upper nodes, ecological in classical 4-motifs
    :param guess_IV_A: value for the number of IV.A or something negative if the
        the number of IV.A-IV.D motifs shall not be computed
    :param array: whether the output should be an array or a dict
    :returns: array or dict with numbers of 4-motifs
    """
    result = {key : 0 for key in sma.MOTIF4_NAMES}
    
    H = G.subgraph(filter(lambda x : G.nodes[x]['sesType'] == level0 or G.nodes[x]['sesType'] == level1, G.nodes))
    data = H.nodes(data=True)
    
    edges_count00 = 0
    edges_count01 = 0
    edges_count11 = 0
    
    classificator = sma.FourMotifClassificator(G)
    for edge in H.edges():
        v1, v2 = edge
        if data[v1]['sesType'] != data[v2]['sesType']:
            edges_count01 += 1
            e1 = v1 if data[v1]['sesType'] == level1 else v2
            s1 = v1 if data[v1]['sesType'] == level0 else v2
            classified = map(classificator,  itertools.product([s1],
                                                               filter(lambda s2: s2 != s1, sma.sesSubgraph(G, level0)),
                                                               [e1],
                                                               filter(lambda e2: e2 != e1, sma.sesSubgraph(G, level1))))
            for cl in classified:
                result[cl] += 1
                del cl
        elif data[v1]['sesType'] == level0:
            edges_count00 += 1
        else:
            edges_count11 += 1
            
    result_array  = numpy.array([result[key] for key in sma.MOTIF4_NAMES])
    # remove symmetry
    result_array1 = numpy.array(numpy.divide(result_array, sma.MOTIF4_SYMMETRIES), dtype = int)
    
    if guess_IV_A >= 0: # guess mode
        V = sma.nodesCount(H)
        expected = numpy.array([scipy.special.comb(V[level1], 2, exact = True) * edges_count00,
                                (V[level0]-1)*(V[level1]-1) * edges_count01,
                                scipy.special.comb(V[level0], 2, exact = True) * edges_count11])
        missing = expected - (result_array1 @ sma.MOTIF4_EDGES)
        totalMotifs = scipy.special.comb(V[level0], 2, exact = True) * scipy.special.comb(V[level1], 2, exact = True)
        n = totalMotifs - sum(result_array1) - guess_IV_A
        v = numpy.array([missing[0], missing[2], n])
        A = numpy.array([[0,-1,1],[-1,0,1],[1,1,-1]])
        
        result_array1[12] = guess_IV_A
        result_array1[13:16] = A @ v
    else:
        result_array1[12:16] = [-1] * 4
    
    if not array:
        return {k : v for k, v in zip(sma.MOTIF4_NAMES, result_array1)}
    return result_array1

def count121MotifsSparse(G : nx.Graph,
                         level0,
                         level1,
                         level2,
                         array : bool = False,
                         optimize_top_adjacent : bool = False,
                         debug_unclassified : bool = False):
    """
    Counts multi level motifs with arities 1, 2, 1, cf. 
    :py:class:`sma.Multi121MotifClassificator`. This function is faster than
    :py:meth:`sma.countMultiMotifs` for sparse graphs, i.e. graphs with few edges.
    The speed up is achieved by (instead of iterating over all suitable tuples of nodes):
        
        - iterating over all nodes in the lowest level
        - for each of them, iterating over the adjacent nodes in the top level
          (or over all nodes from the top level if ``optimize_top_adjacent = False``)
        - for each of them, iterating over all nodes from the middle level that
          are adjacent to the node in the lowest level
          
    With default parameters, this function counts motifs of all classes correctly.
    If ``optimize_top_adjacent = True``, only classes ``3`` and ``4`` are counted correctly.
    
    :param G: the SEN
    :param level0: ``sesType`` of the upper level
    :param level1: ``sesType`` of the middle level
    :param level2: ``sesType`` of the lower level
    :param array: whether the output should be an array or a dict
    :param optimize_top_adjacent: whether the counting should be optimized for
        counting motifs of classes ``3`` and ``4``, see above.
    :param debug_unclassified: whether, in case that ``optimize_top_adjacent = True``,
        the partial count for class ``-1`` should be returned. Per default, the 
        value is set to a negative value.
    """
    motif_info = sma.Motif121Info
    classificator = motif_info.classificator(G)
    result = {k : 0 for k in motif_info.classes}
    for lower in sma.sesSubgraph(G, level2):
        if optimize_top_adjacent:
            top = filter(lambda x: G.nodes[x]['sesType'] == level0, G[lower])
        else:
            top = sma.sesSubgraph(G, level0)
        middle = filter(lambda x: G.nodes[x]['sesType'] == level1, G[lower])
        for a, (b,c) in itertools.product(top,itertools.combinations(middle, 2)):
            cl = classificator((a, b, c, lower))
            result[cl] +=1
    if optimize_top_adjacent:
        result[1] = result[2] = -1 # set to implausible value
        if not(debug_unclassified):
            result[-1] = -1
    else:
        n0 = len(list(sma.sesSubgraph(G, level0)))
        n1 = len(list(sma.sesSubgraph(G, level1)))
        n2 = len(list(sma.sesSubgraph(G, level2)))
        total = n0 * scipy.special.comb(n1, 2, exact = True) * n2
        result[-1] = total - sum(result.values()) + result[-1]
    if array:
        return numpy.fromiter(result.values(), dtype=int)
    return result

def count221MotifsSparse(G : nx.Graph,
                         level0,
                         level1,
                         level2,
                         array : bool = False,
                         optimize_top_adjacent : bool = False,
                         processes : int = 0,
                         chunksize : int = 10000,
                         debug_unclassified : bool = False):
    """
    Counts multi level motifs with arities 2, 2, 1, cf. 
    :py:class:`sma.Multi221MotifClassificator`. This function is faster than
    :py:meth:`sma.countMultiMotifs` for sparse graphs, i.e. graphs with few edges.
    The speed up is achieved by (instead of iterating over all suitable tuples of nodes):
        
        - iterating over all nodes in the lower level
        - for each of them iterating over all pairs of combinations of two nodes
          from the top level, and
        - two nodes from the middle level adjacent to the lower level node.
    
    If ``optimize_top_adjacent = True``, this function iterates only over top level
    nodes which are adjacent to the lower level node. In this case only motifs of
    classes *CLASS*.3 are counted correctly. 
    
    With default parameters, only motifs of classes *CLASS*.2 and *CLASS*.3 are counted
    correctly. The values for ``Unclassified``, *CLASS*.1, *CLASS*.0 will be set 
    to a negative value.
    
    This function supports multiprocessing.
    
    See also :py:class:`sma.DenseMulti221Motifs`, the motif source used here for speed-up.
    
    :param G: the SEN
    :param level0: ``sesType`` of the upper level
    :param level1: ``sesType`` of the middle level
    :param level2: ``sesType`` of the lower level
    :param array: whether the output should be an array or a dict
    :param optimize_top_adjacent: whether the counting should be optimized for
        counting motifs of classes *CLASS*.3, see above.
    :param processes: number of processes for multiprocessing
    :param chunksize: chunksize for multiprocessing
    :param debug_unclassified: whether, the partial count for class ``Unclassified``
        should be returned. Per default, the value is set to a negative value.
    """
    motif_info = sma.Motif221Info
    classificator = motif_info.classificator(G)
    result = {k : 0 for k in motif_info.classes}
    iterator = sma.DenseMulti221Motifs(G, level0, level1, level2, optimize_top_adjacent)
    
    if processes == 0:
        for cl in map(classificator, iterator):
            result[cl] +=1
            del cl
    else:
        with multiprocessing.Pool(processes) as p:
            classified = p.imap_unordered(classificator, iterator, chunksize=chunksize)
            for cl in classified:
                result[cl] +=1
                del cl
            p.close()
            p.join()
    
    # mark implausible results
    if not(debug_unclassified):
        result['Unclassified'] = -1
    implausible = [0,1]
    if optimize_top_adjacent:
        implausible = implausible + [2]
    for v in itertools.product(sma.MOTIF4_NAMES, implausible):
        result['{}.{}'.format(*v)] = -1
        
    # convert to array?
    if array:
        return numpy.fromiter(result.values(), dtype=int)
    return result

class _count222MotifsSparseCounter:
    def __init__(self, G, level0, level1, level2):
        self.G = G
        self.level0 = level0
        self.level1 = level1
        self.level2 = level2
    def __call__(self, lower_edge):
        result = [0] * 3
        lower1, lower2 = lower_edge
        middle_neighbours = set(sma.sesSubgraph(self.G, self.level1, self.G[lower1]))
        top_neighbours    = set(sma.sesSubgraph(self.G, self.level0, self.G[lower1]))
        middles           = middle_neighbours.intersection(set(sma.sesSubgraph(self.G, self.level1, self.G[lower2])))
        tops              = top_neighbours.intersection(set(sma.sesSubgraph(self.G, self.level0, self.G[lower2])))
        for m, t in itertools.product(itertools.combinations(middles, 2), 
                                      itertools.combinations(tops, 2)):
            if (not(self.G.has_edge(*m)) and 
                self.G.has_edge(m[0], t[0]) and
                self.G.has_edge(m[0], t[1]) and
                self.G.has_edge(m[1], t[0]) and
                self.G.has_edge(m[1], t[1])):
                if self.G.has_edge(*t):
                    result[2] += 1
                else:
                    result[1] += 1
            else:
                result[0] += 1
        return result
def count222MotifsSparse(G : nx.Graph,
                         level0,
                         level1, 
                         level2,
                         array : bool = False,
                         processes : int = 0,
                         chunksize : int = 5,
                         debug_unclassified : bool = False):
    """
    Counts multi level motifs with arities 2, 2, 2, cf. 
    :py:class:`sma.Multi222MotifClassificator`. This function is faster than
    :py:meth:`sma.countMultiMotifs` for sparse graphs, i.e. graphs with few edges.
    The speed up is achieved by (instead of iterating over all suitable tuples of nodes):
        
        - iterating over all edges between two lower level nodes
        - for each of them, iterating over all middle level nodes which are simultaenously
          neighbours of both ends of the edge and over all top level nodes which
          are simulatenusly neighbours of both ends of the edge
    
    This function counts only motifs of classes ``3`` and ``4`` correclty. All other
    counts are set to a negative value.
    
    This function supports multiprocessing. Relatively small chunksizes are 
    recommended.
    
    :param G: the SEN
    :param level0: ``sesType`` of the upper level
    :param level1: ``sesType`` of the middle level
    :param level2: ``sesType`` of the lower level
    :param array: whether the output should be an array or a dict
    :param processes: number of processes for multiprocessing
    :param chunksize: chunksize for multiprocessing
    :param debug_unclassified: whether, the partial count for class ``-1``
        should be returned. Per default, the value is set to a negative value.
    """
    result = numpy.zeros((3,), dtype=int)
    lowerlevel = G.subgraph(sma.sesSubgraph(G, level2))
    if processes > 0:
        with multiprocessing.Pool(processes) as p:
            classified = p.imap_unordered(_count222MotifsSparseCounter(G, level0, level1, level2), 
                                          lowerlevel.edges(), chunksize = chunksize)
            for cl in classified:
                result = result + numpy.array(cl)
                del cl
            p.close()
            p.join() 
    else:
        classified = map(_count222MotifsSparseCounter(G, level0, level1, level2), lowerlevel.edges())
        for cl in classified:
                result = result + numpy.array(cl)
                del cl
    real_result = [result[0], -1, -1, result[1], result[2]]
    if not(debug_unclassified):
        real_result[0] = -1
    if not(array):
        return {k : v for k, v in zip([-1, 1, 2, 3, 4], real_result)}
    return numpy.array(real_result)

# Iterators
    
def iterate4Motifs(G : nx.Graph, 
                   level0 = sma.NODE_TYPE_SOC, 
                   level1 = sma.NODE_TYPE_ECO):
    """
    Yields all tuples (s1,s2,e1,e2) where s1,s2 are social nodes,
    e1,e2 are ecological nodes
    
    The tuples are not distict w.r.t. their order, i.e. (1,2) = (2,1).
    The internal order of :py:meth:`sma.sesSubgraph` is kept. 
    
    :param G: the graph
    :param level0: ``sesType`` of the upper nodes, usually :py:const:`sma.NODE_TYPE_SOC`
    :param level1: ``sesType`` of the lower nodes, usually :py:const:`sma.NODE_TYPE_ECO`
    """
    socialSubgraph = sesSubgraph(G, level0)
    ecologicalSubgraph = sesSubgraph(G, level1)
    return itertools.starmap(lambda x,y: (*x,*y),itertools.product(
            itertools.combinations(socialSubgraph, 2),
            itertools.combinations(ecologicalSubgraph, 2)))

def iterate3Motifs(G : nx.Graph, level0, level1):
    """
    Yields all tuples (a, b1, b2) where a is a node from ``level0`` and b1, b2 
    nodes from ``level1``. This is a backend function for :py:meth:`sma.iterate3EMotifs`
    and :py:meth:`sma.iterate3SMotifs`.
    
    :param G: the graph
    :param level0: ``sesType`` of the distinct node, e.g. in 3E-motifs :py:const:`sma.NODE_TYPE_ECO`
    :param level1: ``sesType`` of the other nodes, e.g. in 3E-motifs :py:const:`sma.NODE_TYPE_SOC`
    """
    ecologicalSubgraph = sesSubgraph(G, level0)
    socialSubgraph = sesSubgraph(G, level1)
    return map(lambda x: (x[0],*x[1]), 
               itertools.product(ecologicalSubgraph, 
                                 itertools.combinations(socialSubgraph, 2)))

def iterate3EMotifs(G : nx.Graph,
                    level0 = sma.NODE_TYPE_ECO,
                    level1 = sma.NODE_TYPE_SOC):
    """
    Yields all tuples (e, s1, s2) where e is an ecological node and
    s1,s2 are social nodes.
    
    The tuples are not distict w.r.t. their order, i.e. (1,2) = (2,1).
    The internal order of :py:meth:`sma.sesSubgraph` is kept.
    
    See also :py:meth:`sma.iterate3Motifs`.
    
    :param G: the graph
    :param level0: ``sesType`` of the distinct node, here :py:const:`sma.NODE_TYPE_ECO`
    :param level1: ``sesType`` of the other nodes, here :py:const:`sma.NODE_TYPE_SOC`
    """
    return iterate3Motifs(G, level0, level1)

def iterate3SMotifs(G : nx.Graph,
                    level0 = sma.NODE_TYPE_SOC,
                    level1 = sma.NODE_TYPE_ECO):
    """
    Yields all tuples (s, e1, e2) where s is a social node and 
    e1, e2 are ecological nodes.
    
    The tuples are not distict w.r.t. their order, i.e. (1,2) = (2,1).
    The internal order of :py:meth:`sma.sesSubgraph` is kept.
    
    See also :py:meth:`sma.iterate3Motifs`.
    
    :param G: the graph
    :param level0: ``sesType`` of the distinct node, here :py:const:`sma.NODE_TYPE_SOC`
    :param level1: ``sesType`` of the other nodes, here :py:const:`sma.NODE_TYPE_ECO`
    """
    return iterate3Motifs(G, level0, level1)

def sesSubgraph(G : nx.Graph, sesType : int, source = None) -> filter:
    """
    Filters only social or ecological nodes from a given graph. The returned 
    object is a filter. If you need a :py:class:`networkx.Graph`, you may want
    to call ``G.subgraph`` on the result.
    
    .. code :: Python
    
        # Let G be some SEN
        socialNodes = list(sma.sesSubgraph(G, sma.NODE_TYPE_SOC))
        socialGraph = G.subgraph(sma.sesSubgraph(G, sma.NODE_TYPE_SOC))
    
    :param G: the graph
    :param sesType: the type (either social or ecological), see
        :py:const:`sma.NODE_TYPE_ECO` and :py:const:`sma.NODE_TYPE_SOC`.
    :param source: an alternative source for nodes, if ``None``, ``G.nodes`` is used
    :returns: filter object with yields the nodes. Be aware that you will need to 
        call :py:func:`list` on the result to obtain a list.
    """
    if source == None:
        source = G.nodes
    return filter(lambda x : G.nodes[x]['sesType'] == sesType, source)

# Simple properties

def density(G : nx.Graph) -> float:
    """
    Returns the density of a given graph G = (V, E) defined as 
    
    .. math ::
        |E| / (|V| (|V|-1) / 2).
    
    See also :py:meth:`sma.densityMatrix`.
    
    :param G: the SEN
    """
    E = G.number_of_edges()
    V = G.number_of_nodes()
    return 2 * E / V / (V-1)

def densityMatrix(G : nx.Graph) -> numpy.ndarray:
    """
    Returns a matrix with edge densities for every subsystem, i.e. the quotient 
    of :py:meth:`sma.edgesCountMatrix` and :py:meth:`sma.maxEdgeCountMatrix`.
    
    See also :py:meth:`sma.density`.
    
    :param G: the SEN
    """
    v = sma.edgesCountMatrix(G)
    m = sma.maxEdgeCountMatrix(sma.nodesCount(G, array=True))
    return numpy.divide(v, m, where = (m!=0))

def nodesCount(G : nx.Graph, array : bool =False) -> numpy.ndarray:
    """
    Returns either a dict or an array with the number of nodes in each ``sesType``.
    
    :param G: the SEN
    :param array: whether the result is an array or a dict
    :raises TypeError: If one or more nodes have no sesType
    """
    l = list(dict(G.nodes(data='sesType')).values())
    if None in l:
        raise TypeError('one or more nodes have no sesType, use sma.untypedNodes() to find them')
    types, counts = numpy.unique(l, return_counts = True)
    if array:
        return [counts[types == k][0] if k in types else 0 for k in range(max(types)+1)]
    else:
        return {k : counts[types == k][0] if k in types else 0 for k in range(max(types)+1)}

def untypedNodes(G : nx.Graph) -> map:
    """
    Returns a map object yielding all nodes which do not have a ``sesType``-attribute.
    Most functions used in analyses require all nodes to have such an attribute.
    
    Example: List all untyped nodes:
    
    .. code :: Python
    
        # Let G be a SEN
        print(list(sma.untypedNodes(G)))
    
    :param G: the SEN
    :returns: map object yielding all untyped nodes    
    """
    return map(lambda v : v[0],
               filter(lambda v : v[1] is None, 
                      dict(G.nodes(data='sesType')).items()))

def edgesCount(G : nx.Graph, array : bool =False) -> dict:
    """
    Classifies all edges in a SEN according to the following classes.
    
        - :py:const:`sma.EDGE_TYPE_SOC_SOC` for edges linking two social nodes
        - :py:const:`sma.EDGE_TYPE_ECO_ECO` for edges linking two ecological nodes
        - :py:const:`sma.EDGE_TYPE_ECO_SOC` for edges linking a social and an ecological node
    
    Since edges are undirected no other types occur. 
    
    See :py:meth:`sma.edgesCountMatrix` for multi-level support.
    
    :param G: SEN as networkx graph
    :param array: whether the result is an array or a dict
    :returns: dict containing type-count pairs according to the list of types above
    """
    def _classify(edge):
        v1, v2 = edge
        types = [G.nodes[v1]['sesType'], G.nodes[v2]['sesType']]
        if types == [sma.NODE_TYPE_SOC, sma.NODE_TYPE_SOC]:
            return sma.EDGE_TYPE_SOC_SOC
        if types == [sma.NODE_TYPE_ECO, sma.NODE_TYPE_ECO]:
            return sma.EDGE_TYPE_ECO_ECO
        if sma.NODE_TYPE_ECO in types and sma.NODE_TYPE_SOC in types:
            return sma.EDGE_TYPE_ECO_SOC
        raise TypeError('sma.edgesCount supports only two-level networks, use sma.edgesCountMatrix instead')
    classification = list(map(_classify, G.edges))
    types, counts = numpy.unique(classification, return_counts=True)
    if array:
        return [int(counts[types == k][0]) if k in types else 0 for k in [sma.EDGE_TYPE_SOC_SOC, sma.EDGE_TYPE_ECO_SOC, sma.EDGE_TYPE_ECO_ECO] ]
    else:
        return { k : int(counts[types == k][0]) if k in types else 0 for k in [sma.EDGE_TYPE_SOC_SOC, sma.EDGE_TYPE_ECO_SOC, sma.EDGE_TYPE_ECO_ECO] }

def edgesCountMatrix(G : nx.Graph, nTypes : int = None) -> numpy.ndarray:
    """
    This function provides an analogous result as :py:meth:`sma.edgesCount` for
    multi-level graphs. For a SEN with :math:`n` types of nodes the returned matrix
    is upper-triangular and of dimension :math:`n \\times n`. The entry :math:`(i,j)`
    contains the number of edges in the given graph linking nodes of type :math:`i`
    and :math:`j`.

    The sum over all entries equals the total number of edges in the SEN.

    The returned matrix can be used as a parameter for :py:meth:`sma.randomMultiSENs`.

    :param G: a SEN, possibly with more than two levels
    :param nTypes: the number of types in the SEN. An ordinary social-ecological network
        has two levels. If ``None``, the number of levels is determined automatically
        using :py:meth:`sma.sesTypes`.
    """
    if nTypes is None:
        nTypes = max(sma.sesTypes(G)) + 1
    res = numpy.zeros((nTypes, nTypes), dtype=int)
    for v1, v2 in G.edges:
        res[G.nodes[v1]['sesType'], G.nodes[v2]['sesType']]+=1
    return numpy.triu(res + res.T) - numpy.diag(numpy.diag(res))

def adjacentEdgesCount(G : nx.Graph, node, nTypes = 2) -> numpy.ndarray:
    """
    The :math:`i`-th entry in the returned matrix row is the number of neighbors
    the given node has of the type :math:`i`. For example, in a ordinary two-level
    SEN with social and ecological nodes the first entry contains the number of 
    adjacent ecological nodes (type 0) whereas the second entry contains the number
    of social nodes (type 1). Multi-level SENs are supported.
    
    :param G: the SEN
    :param node: the node
    :param nTypes: number of types in the SEN, usually 2 (social and ecological)
    :returns: matrix row with numbers of adjacent vertices by type
    """
    neighbors = G[node]
    classification = list(map(lambda n : G.nodes[n]['sesType'], neighbors))
    types, counts = numpy.unique(classification, return_counts = True)
    return numpy.array([counts[types == i][0] if i in types else 0 for i in range(nTypes)], dtype=int)

def total4Motifs(G : nx.Graph) -> int:
    """
    Returns the total number of 4-motifs in a graph, i.e.
    
    .. math ::
        
        {n_s \\choose 2} {n_e \\choose 2} = \\frac{n_s(n_s-1)n_e(n_e-1)}{4}
        
    where n_s and n_e are respectively the numbers of social and ecological nodes.
    
    See also :py:meth:`sma.totalMultiMotifs`.
    
    :param G: the graph
    :returns: total number of 4-motifs
    """
    return totalMultiMotifs(G, 2, 2)

def total3EMotifs(G : nx.Graph) -> int:
    """
    Returns the total number of 3E-motifs in a graph, i.e.
    
    .. math ::
        
        {n_s \\choose 2} n_e = \\frac{n_s(n_s-1)n_e}{2}
        
    where n_s and n_e are respectively the numbers of social and ecological nodes.
    
    See also :py:meth:`sma.totalMultiMotifs`.
    
    :param G: the graph
    :returns: total number of 3E-motifs
    """
    return totalMultiMotifs(G, 1, 2)

def total3SMotifs(G : nx.Graph) -> int:
    """
    Returns the total number of 3S-motifs in a graph, i.e.
    
    .. math ::
        
        {n_e \\choose 2} n_s = \\frac{n_e(n_e-1)n_s}{2}
        
    where n_s and n_e are respectively the numbers of social and ecological nodes.
    
    See also :py:meth:`sma.totalMultiMotifs`.
    
    :param G: the graph
    :returns: total number of 3S-motifs
    """
    return totalMultiMotifs(G, 2, 1)

def totalMultiMotifs(G : nx.Graph, *arities : int) -> int:
    """
    Returns the total number of multilevel motifs of given aritíes (number of
    nodes taken from each level) for the given graph. If the given arities are
    :math:`a_1, \dots, a_n`, then the returned value is
    
    .. math ::
        
        \prod_{i=1}^n \\binom{N_i}{a_i}
        
    where :math:`N_1, \dots, N_n` are the amounts of vertices in the different levels.
    
    The following function calls are equivalent, cf. :py:meth:`sma.total4Motifs`
    but also :py:meth:`sma.total3EMotifs` and :py:meth:`sma.total3SMotifs`:
        
    .. code :: Python
        
        sma.total4Motifs(G, multi = True)
        # and
        sma.totalMultiMotifs(G, 2, 2)
    
    :param G: the SEN
    :param arities: the arities, i.e. number of nodes taken from each level
    """
    total = 1
    for level, arity in enumerate(arities):
        if arity >= 1:
            N = len(list(sma.analyse.sesSubgraph(G, level)))
            total *= scipy.special.comb(N, arity, exact = True)
    return total

def sesTypes(G : nx.Graph) -> set:
    """
    Returns the set of all ``sesTypes`` occurring in the given graph.
    
    :param G: the SEN
    :returns: set of all ``sesTypes``, e.g. :math:`\{0, 1\}` in standard two-level
        SENs
    """
    return set(map(lambda k : k[1]['sesType'], G.nodes.data(data=True)))


#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import networkx as nx
import itertools, numpy
import sma
import multiprocessing
import operator

def cooccurrenceTable(G : nx.Graph, *motifs, to_array : bool = False, to_dict : bool = False):
    """
    Returns the co-occurrence table for a given set of motif generators. This table
    contains one row for each node in the given SEN G. The :math:`i`-th entry for 
    the :math:`j`-th row represents the number of occurences of the :math:`j`-th 
    node in motifs yielded by the :math:`i`-th motif generator.
    
    For example, one could want to count the number of occurences of nodes in closed
    and open triangles::
        
        import sma
        # let G be some SEN
        result = sma.cooccurrenceTable(G, 
                                       sma.ThreeEMotifs(G) & sma.is3Class('I.C'), 
                                       sma.ThreeEMotifs(G) & sma.is3Class('II.C'))
    
    Then the first entry for each node contains the number of open triangles 
    (class I.C motif) this node is involved in and the second entry the number of
    closed triangles (class II.C motif). Note, that in this example the involvement
    of a node at any position (not only at the distinct position) counts, cf.
    :py:meth:`sma.triangleCoefficient`.
    
    See also :py:meth:`sma.cooccurrenceTableFull`.
    
    :param G: the SEN
    :param motifs: a list of motif generators, cf. :py:class:`sma.MotifIterator`
    :param to_array: boolean indicating whether the output should be a numpy
        array, entries for nodes a given in the order returned by the graph
    :param to_dict: boolean indicating whether the output should be a two-level
        dict. Values can be accessed by ``result[node][motif]`` where motif is 
        the string representation of the motif
    :returns: co-occurence table as described above
    """
    result = {node : [0]*len(motifs) for node in G.nodes}
    for generator, i in zip(motifs, itertools.count()):
        for motif in generator:
            for node in motif:
                result[node][i] += 1
    if to_array:
        return numpy.array(list(result.values()))
    if to_dict:
        gstrings = list(map(lambda x : x.__str__(), motifs))
        return {node : {gstrings[i] : result[node][i] for i in range(len(motifs))} for node in G.nodes}
    return result

def cooccurrenceTableFull(G : nx.Graph, 
                          iterator : sma.MotifIterator, 
                          classificator : sma.MotifClassificator,
                          to_array : bool = False):
    """
    This function returns a dict/array similar to the result of :py:meth:`sma.cooccurrenceTable`.
    This function should be used when the values for several classes of motifs taken
    from the same source iterator are of interest. For example, if all 3-motifs
    shall be fully classified, :py:meth:`sma.cooccurrenceTableFull` should be 
    preferred over :py:meth:`sma.cooccurrenceTable` since it does not incur any 
    redundant costs.
    
    If the parameter ``to_array`` is set to ``False`` (default), the result is 
    a dictionary which maps vertices to subdictionaries mapping motif classes (taken
    from the classificator) to the integer representing the number of occasions this
    vertic occurs in a motif of this class. If ``to_array`` is flipped, the result 
    is converted to a matrix with one row for each vertex and one column for each
    motif class.
    
    The sum of all entries equals the total number of motifs in the iterator
    multiplied by the arity of the classificator.
    
    :param G: a SEN
    :param iterator: a :py:class:`sma.MotifIterator` as source of motifs
    :param classificator: a :py:class:`sma.MotifClassificator` for classifying the
        motifs
    :param to_array: whether the result should be an array or a dict.
    :returns: cooccurrences table featuring values for all classes of motifs
    
    """
    result = {node : {} for node in G.nodes}
    for motif in iterator:
        for node in motif:
            typ = classificator(motif)
            if typ in result[node]:
                result[node][typ] += 1
            else:
                result[node][typ] = 1
    if to_array:
        return numpy.array([[row[motif] if motif in row else 0 for motif in classificator.info().classes] for row in result.values()])
    return result
    
def motifMultigraph(G : nx.Graph, *motifs, attr = 'motif') -> nx.MultiGraph:
    """
    Returns a multigraph of motifs. In this multigraph every node represents motif.
    For example, when the second parameter is ``sma.FourMotif(G)``, then every
    node in the returned graph represents one 4-motif. Each edge connecting two 
    nodes :math:`m_1`, :math:`m_2` represents one node in the original graph that is shared by two
    motifs that :math:`m_1` and :math:`m_2` represent.
    
    Note that the motif multigraph does not contain any loops since each motifs
    shares trivially all its vertices with itself.
    
    A list of motif iterators must be given, cf. :py:class:`sma.MotifIterator`.
    In this way different classes of motifs can be incorporated in one multigraph.
    For example, one could be interested in the adjacence of open triangles with
    distinct social node (social motif I.C) and closed triangles with distinct 
    ecological node (ecological motif II.C). Then these two motif iterators can
    be provided as input, cf. :py:class:`sma.ThreeEMotifs`, :py:class:`sma.ThreeSMotifs`
    and :py:class:`sma.is3Class`.
    
    The multigraph can be converted to a weighted simple graph with edge weights
    corresponding to edge multiplicities in the multigraph using :py:meth:`sma.multiToWeightedGraph`.
    See also :py:meth:`sma.motifWeightedGraph`.
    
    The class of the motifs, i.e. a string representation of the given motif 
    iterators, is stored as nodal attribute.
    
    :param G: a SEN
    :param motifs: one or several instances of :py:class:`sma.MotifIterator`
    :param attr: attribute key for the nodal attribute representing the class
        of the motifs, default ``motif``
    :returns: motif multigraph
    """
    M = nx.MultiGraph()
    for motif in motifs:
        M.add_nodes_from(motif, **{attr : motif.__str__()})
    sesTypes = nx.get_node_attributes(G, 'sesType')
    for m1, m2 in itertools.combinations(M.nodes, 2):
        common = set(m1).intersection(set(m2))
        for n in common:
            M.add_edge(m1, m2, sesType=sesTypes[n], name=n)
    return M

def motifWeightedGraph(G : nx.Graph, *motifs, attr='weight') -> nx.Graph:
    """
    Returns a weighted graph corresponding to the motif multigraph generated 
    by :py:meth:`sma.motifMultigraph`. That two motifs share vertices is not
    encoded by multiple edges but by edge weights.
    
    See :py:meth:`sma.multiToWeightedGraph`, the function that takes care of the
    translation.
    
    :param G: the SEN
    :param motifs: one or several instances of :py:class:`sma.MotifIterator`
    :param attr: the attribute key for the weight attribute, default is ``weight``.
    """
    M = motifMultigraph(G, *motifs)
    return sma.multiToWeightedGraph(M, attr)

def motifClassMatrix(G : nx.Graph, 
                     iterator : sma.MotifIterator, 
                     classificator : sma.MotifClassificator,
                     as_symmetric : bool = False) -> numpy.ndarray:
    """
    In the returned quadratic matrix every column and every row represents a motif
    class, i.e. a class of motif as recorgnized by the given :py:class:`sma.MotifClassificator`.
    For example, in case of 3-motifs the first column and the first row represent
    class 'I.A' motifs whereas the last row / column represents 'II.C' motifs. In
    total, the matrix would be of dimension :math:`6 \\times 6` since there are six
    3-motifs.
    
    The :math:`(i,j)`-th entry represents the number of vertices in the given SEN shared
    by motifs of class :math:`i` and :math:`j`. If a vertex is contained in several class 
    :math:`i` and class :math:`j` motifs, it is counted multiply.
    
    Let :math:`(C_{ij})` denote the returned matrix. Then for the sum of all upper
    triangular entries the following correspondance holds:
    
    .. math ::
        
        \\sum_{i \\geq j} C_{ij} = \\frac12 \\sum_{M_1} \\sum_{M_2 \\neq M_1} 
        \\left| M_1 \\cap M_2 \\right|
        
    where :math:`M_1` and :math:`M_2` are motifs taken from the iterator.
    
    This function computes the desired matrix based on the result of :py:meth:`sma.cooccurrenceTableFull`.
    For every vertex its occurrence in each of the possible classes of motifs is
    counted. Let red, blue and orange be three distinct motif classes. If a vertex occurres 
    :math:`n` times in red motifs, :math:`m_1` times in blue motifs and :math:`m_2`
    times in orange motifs, then it establishes :math:`\\binom{n}{2}` connections
    between red motifs. Hence it contributes this number to the red diagonal entry.
    This is the number of edges in the complete graph :math:`K_n`. Moreover, the
    vertex establishes :math:`m_1 \\cdot m_2` connections between blue and orange
    motifs. This is the number of edges in the complete bipartite graph :math:`K_{m_1,m_2}`
    and the vertex' contribution to the blue-orange off-diagonal entry.
    
    :param G: the SEN
    :param iterator: a :py:class:`sma.MotifIterator` as a source of motifs
    :param classificiator: a :py:class:`sma.MotifClassificator` for classifying 
        the motifs. Note that it must match with the given iterator.
    :param as_symmetric: per default the returned matrix is upper triangular. If
        this switch is set to ``True`` the returned matrix will be symmetrical, with
        the upper entries copied to the matrix' lower half.
    """
    ctable = cooccurrenceTableFull(G, iterator, classificator, to_array = True)
    choose2 = lambda n : n*(n-1)//2
    diag = numpy.diag(sum(choose2(ctable)))
    offdiag = numpy.array([[numpy.dot(ctable[:,i],ctable[:,j]) if i < j else 0 
                            for j in range(len(classificator.info().classes))] 
                            for i in range(len(classificator.info().classes))])
    if as_symmetric:
        return offdiag + diag + numpy.transpose(offdiag)
    else:
        return offdiag + diag
 
def motifClassGraph(G : nx.Graph, 
                    iterator : sma.MotifIterator, 
                    classificator : sma.MotifClassificator) -> nx.Graph:
    """
    Returns an :py:class:`networkx.Graph` with one node for every motif class as 
    described by the given :py:class:`sma.MotifClassificator`. The edges are weighted
    by the entries of :py:meth:`sma.motifClassMatrix`. The graph contains loops.
    
    :param G: the SEN
    :param iterator: a :py:class:`sma.MotifIterator` as a source of motifs
    :param classificiator: a :py:class:`sma.MotifClassificator` for classifying 
        the motifs. Note that it must match with the given iterator.
    """
    graph = nx.from_numpy_array(motifClassMatrix(G, iterator, classificator, as_symmetric=True))
    return nx.relabel_nodes(graph, dict(zip(itertools.count(), classificator.info().classes)), copy=True)

class _cooccurrenceEdgeTableFullMapper:
    def __init__(self, classificator, nodes_rows, nodes_columns, dyad):
        self.classificator = classificator
        self.indexer_rows    = {n : i for n, i in zip(nodes_rows, itertools.count())}
        self.indexer_columns = {n : i for n, i in zip(nodes_columns, itertools.count())}
        self.indexer_motifs  = {n : i for n, i in zip(classificator.info().classes, itertools.count())}
        self.dyad = dyad
    def __call__(self, motif):
        cl = self.classificator(motif)
        if type(self.dyad) is int:
            return (self.indexer_rows[motif[self.dyad][0]], 
                    self.indexer_columns[motif[self.dyad][1]], 
                    self.indexer_motifs[cl]) 
        else:
            return (self.indexer_rows[motif[self.dyad[0]]], 
                    self.indexer_columns[motif[self.dyad[1]]], 
                    self.indexer_motifs[cl])


def cooccurrenceEdgeTableFull(G : nx.Graph,
                              iterator : sma.MotifIterator,
                              classificator : sma.MotifClassificator,
                              dyad,
                              levels : tuple,
                              processes : int = 0,
                              chunksize : int = 10000):
    """
    Computes a cooccurrence table on edge level. Given an edge :math:`(v_1, v_2)`
    of a motif, :math:`v_1` in level :math:`i`, :math:`v_2` in level :math:`j`, this
    table consists of :math:`|V_i| \\times |V_j| \\times M` entries where :math:`V_i`,
    :math:`V_j` denotes the set of notes from level :math:`i`, resp. :math:`j` in
    the SEN and :math:`M` denotes the number of motif classes as classified by the
    given classificator. The :math:`(a,b,c)`th entry contains the number of times
    the edge :math:`(a,b)` occurs in a motif of class :math:`c`.
    
    The edge is specified using the ``dyad`` parameter. This parameter must contain
    a tuple of two integers specifying the index of the nodes spanning the edge in 
    the motifs provided by the iterator. For example, a typical 3E-motif looks like
    :math:`(e, s_1, s_2)` where :math:`s_1` and :math:`s_2` are social and :math:`e`
    is ecological. In this case, ``dyad = (1,2)`` would imply that the cooccurrence
    matrix for the edge :math:`(s_1, s_2)` would be computed. For technical reasons,
    a paramter ``levels`` must be provided. ``levels`` must be a tuple of length two
    specifying the ``sesType`` of the nodes in edge specified by ``dyad``. In the
    example, ``levels`` would be (:py:const:`sma.NODE_TYPE_SOC`,:py:const:`sma.NODE_TYPE_SOC`).
    
    Multiprocessing is supported. Use parameters ``processes`` and ``types``.
    
    **Example** Compute the cooccurrence table for 3S-motifs
    
    .. code :: Python 
    
        sma.cooccurrenceEdgeTableFull(G, 
                                      sma.ThreeSMotifs(G), 
                                      sma.ThreeMotifClassificator(G), 
                                      (1,2), # dyad, social nodes
                                      (0,0)) # levels, both social
    
    :param G: the SEN
    :param iterator: source of motifs
    :param classificator: classificator for the motifs
    :param dyad: specificiation of the esge for which the cooccurrence table is 
        computed (tuple of length 2)
    :param levels: ``sesTypes`` of the nodes in ``dyad`` (tuple of length 2)
    :param processes: number of processes for multiprocessing
    :param chunksize: chunksize for multiprocessing
    :returns: three values: the cooccurrence table, list of nodes as index for the rows,
        list of nodes as index for the columns
    """
    nodes_rows = list(sma.sesSubgraph(G, levels[0]))
    if levels[0] == levels[1]:
        nodes_columns = nodes_rows
    else:
        nodes_columns = list(sma.sesSubgraph(G, levels[0]))
    
    matrix = numpy.zeros((len(nodes_rows), len(nodes_columns), len(classificator.info().classes)), dtype=int)
    mapper = _cooccurrenceEdgeTableFullMapper(classificator, 
                                              nodes_rows, 
                                              nodes_columns, 
                                              dyad)
    
    if processes == 0:
        for index in map(mapper, iterator):
            matrix[index] +=1
            del index
    else:
        with multiprocessing.Pool(processes) as p:
            classified = p.imap_unordered(mapper, iterator, chunksize = chunksize)
            for index in classified:
                matrix[index] +=1
                del index
            p.close()
            p.join()
    
    if levels[0] == levels[1]:
        for i in range(len(classificator.info().classes)):
            matrix[:,:,i] = numpy.triu(matrix[:,:,i] + matrix[:,:,i].T) - numpy.diag(numpy.diag(matrix[:,:,i]))
    
    return matrix, nodes_rows, nodes_columns

def edgeContributionList(G : nx.Graph, *arities, roles = [], level : int) -> numpy.ndarray:
    """
    Computes edge contribution lists for a certain family of motifs. This list
    contains one entry for every possible edge in the level of the network specified
    by ``level``. The entry associated with a dyad is number of times this dyad
    occurs in a certain family of motifs. The dyad itself is not taken into account
    here. In this way, open and closed motifs (i.e. motifs with a dyad in level
    ``level`` and otherwise equivalent motifs without this dyad) are grouped together.
    The value measures the contribution of a dyad in level ``level`` to open/closed
    motifs of similar class.
    
    This only makes sense if the motif specified by ``arities`` (and ``roles``)
    contains two nodes in level ``level``.
    
    For example, when working with 3-motifs (and ``level`` equalling the ``sesType``
    of the non-distinct nodes), motifs I.A and II.A, I.B and II.B and I.C and II.C
    are grouped together.
    
    :param G: the SEN
    :param arities: arities of the (multi-level) motif
    :param roles: optional roles for position matching, cf. :py:meth:`sma.matchPositions`
    :param level: level as described above
    :returns: tuple of:
        
        - two-dimensional array with one row for every group of motif (grouped
          by open/closed) and one column for every dyad on level ``level``
        - iterator giving the edges which index the columns of the array, call 
          ``list()`` on it to obtain a list of indices
    
    :raises NotImplementedError: if the motif signature is not supported
    """
    signature  = sma.multiSignature(arities)
    motif_info = sma.motifInfo(signature)
    positions  = sma.matchPositions(motif_info.signature, arities, roles)
    assert level in positions, 'selected motif does not reach selected level'
    index = positions.index(level)
    assert motif_info.signature[index] == 2, 'motif does not contain 2 vertices on the requested level'
    offset = sum(motif_info.signature[:index])

    iterator   = sma.MultiMotifsNormalized(G, *arities, roles = roles, motif_info = motif_info)
    classificator = motif_info.classificator(G)
    
    table, nodes, _ = sma.cooccurrenceEdgeTableFull(G,
                                                    iterator,
                                                    classificator,
                                                    (offset, offset+1),
                                                    (level, level))

    per_motif = numpy.array([table[:,:,i][numpy.triu_indices(table.shape[0],1)] for i in range(table.shape[2])])
    
    # group together
    indexer = {motif : i for i, motif in enumerate(motif_info.classes)}
    abstract_level = positions.index(level)
    if abstract_level not in motif_info.projections:
        raise NotImplementedError('ActorsChoice not implemented for signature %s on level %d' \
                                  % (str(motif_info.signature), level))
    grouper = motif_info.projections[abstract_level]
    result  = numpy.array([sum(per_motif[indexer[gr]] for gr in group) for group in grouper])
    
    edges = itertools.combinations(nodes, 2)
    return result, edges

def identifyGaps(G : nx.Graph, 
                 motif_identifier : str, 
                 level : int = -1) -> list:
    """
    Computes a list of gaps (specific dyads) together with their contribution to 
    the number of motifs of a specific class.
    
    The contribution of a gap is the number of motifs of a fixed class which would
    be created by replacing the gap with an edge. This function computes this contribution
    for every gap on the specified level.
    
    The behaviour above applies when the specified motif contains the edge on the
    given level (i.e. the motif is a *closed motif*). If the edge is not present
    in the specified motif (the motif is *open*), the contribution is definied the
    number of motifs of the specified class are created by removing an edge from the
    network.
    
    :param G: the SEN
    :param motif_identifier: motif identifier string of the motif
    :param level: dyads of which level (``sesType``) are to be considered
    :returns: list of tuples ordered by the second entry in descending order
        consisting of
    
        - a tuple specifying an edge
        - the contribution of this edge to the number of motifs of the given type
          when being flipped
    
    """
    arities, roles, motif = sma.parseMotifIdentifier(motif_identifier)
    signature  = sma.multiSignature(arities)
    motif_info = sma.motifInfo(signature)
    if level < 0:
        assert 2 in motif_info.signature, 'signature does not contain two'
        positions  = sma.matchPositions(motif_info.signature, arities, roles)
        level = positions[motif_info.signature.index(2)]
    contributions, edges = sma.edgeContributionList(G, 
                                                    *motif_info.signature, 
                                                    roles = positions, 
                                                    level = level)
    abstract_level = positions.index(level)
    grouper = motif_info.projections[abstract_level]
    index, group = next(filter(lambda x: motif in x[1], enumerate(grouper)))
    # we want to identify gaps, so ties which would close the largest number of gaps,
    # i.e. whose presence would increase the number of motifs specified by motif_idenitifer
    row = contributions[index]
    values = zip(edges, row)
    if group[0] == motif:
        # motif is an open motif
        values_filtered = filter(lambda x : G.has_edge(*x[0]), values)
    else:
        # motif is a closed motif
        values_filtered = filter(lambda x : not(G.has_edge(*x[0])), values)
    return sorted(values_filtered, key=operator.itemgetter(1), reverse=True)

def isClosedMotif(motif_identifier : str, level : int = -1) -> bool:
    """
    Queries the motif database and determines whether a motif is closed or open
    with respect to a given level.
    
    See :py:meth:`sma.identifyGaps`.
    
    :param motif_identifier: motif identifier string
    :param level: level (``sesType``) for the gap
    :returns: true/false whether the motif is closed.
    :raises NotImplementedError: if the motif database does not know..
    """
    arities, roles, motif = sma.parseMotifIdentifier(motif_identifier)
    signature             = sma.multiSignature(arities)
    motif_info            = sma.motifInfo(signature)
    if level >= 0:
        positions         = sma.matchPositions(motif_info.signature, arities, roles)
        abstract_level    = positions.index(level)
    else:
        abstract_level    = motif_info.signature.index(2)
    if abstract_level not in motif_info.projections:
        raise NotImplementedError('closed/open relations not implemented for motifs with signature %s on abstract level %d'\
                                  % (str(motif_info.signature), abstract_level))
    for o, c in motif_info.projections[abstract_level]:
        if c == motif:
            return True
        if o == motif:
            return False
    raise TypeError('unknown motif %s' % motif)
    


import networkx as nx
import sma
import numpy

def _nodeColors(G : nx.Graph, color_map):
    return list(map(lambda x : color_map[x[1]['sesType']], G.nodes(data=True)))

def drawSEN(G : nx.Graph, color_map = sma.COLORS_TYPES, **kwargs) :
    """
    Draws a social-ecological network using :py:meth:`networkx.draw_kamada_kawai`.
    The default drawing behaviour can be overwritten by providing the parameter ``pos``.
    
    :param G: the graph
    :param color_map: a dict mapping the ``sesTypes`` of the nodes to colors,
        default is :py:const:`sma.COLORS_TYPES`.
    :param kwargs: parameters for :py:meth:`networkx.draw_kamada_kawai`
    """
    if 'pos' in kwargs:
        nx.draw(G, 
                  with_labels=True, 
                  node_color=_nodeColors(G, color_map),
                  **kwargs)
    else:
        nx.draw_kamada_kawai(G, 
                      with_labels=True, 
                      node_color=_nodeColors(G, color_map),
                      **kwargs
                      )

def drawGeoSEN(G : nx.Graph, 
               longAttribute : str = 'long', 
               latAttribute : str = 'lat', 
               **kwargs):
    """
    Draws a SEN whose nodes have a location property. :py:meth:`draw_networkx`
    is used as backend.
    
    :param G: the graph
    :param longAttribute: the name of the nodal attribute which contains the longitude
        of the nodes
    :param latAttribute: the name of the nodal attribute which contains the latitude
        of the nodes
    :param kwargs: parameters for :py:meth:`draw_networkx`
    """
    positions = {n : (d[longAttribute], d[latAttribute]) for n, d in G.nodes(data=True)}
    nx.draw_networkx(G, pos=positions, node_color = _nodeColors(G), **kwargs)

def drawMotif(G : nx.Graph, motif : tuple, **kwargs):
    """
    Draws a specific motif in a SEN. 
    
    Example:
    
    .. code :: Python
        
        # let G be some SEN
        motif = sma.exemplifyMotif(G, '1,2[II.C]')
        sma.drawMotif(G, motif)
    
    :param G: the SEN
    :param motif: a motif (tuple of nodes)
    :param kwargs: more parameters for :py:meth:`sma.drawSEN` and 
        :py:meth:`networkx.draw_networkx`.
    """
    H = G.subgraph(motif)
    drawSEN(H, **kwargs)

def layer_layout(G : nx.Graph, scale: float = 5):
    """
    Computes a layout for a SEN which places the nodes of one type on one horizontal
    line. Use it together with :py:meth:`sma.drawSEN`.
    
    .. code-block :: Python
        
        import sma
        # Let G be some SEN
        sma.drawSEN(G, pos=sma.layer_layout(G))
    
    :param G: the SEN
    :param scale: scaling factor
    :returns: layout for :py:meth:`sma.drawSEN`
    """
    nodesCounts = sma.nodesCount(G, array=False)
    maxType = max(nodesCounts.keys())
    source = {k : iter(list(numpy.linspace(0, scale, v))) for k, v in nodesCounts.items()}
    return {node : (next(source[typ]), maxType-typ) for node, typ in G.nodes(data = 'sesType')}

def advanced_layer_layout(G : nx.Graph, space : float = 1):
    """
    Computes a layout for a SEN which places the nodes of one type close to each
    other. Use it together with :py:meth:`sma.drawSEN`.
    
    :param G: the SEN
    :param space: space between the layers
    :returns: layout for :py:meth:`sma.drawSEN`
    """
    nodesCounts = sma.nodesCount(G, array=False)
    shift = numpy.zeros((2))
    positions = {}
    for k, v in nodesCounts.items():
        subgraph = G.subgraph(sma.sesSubgraph(G, k))
        pos = nx.spring_layout(subgraph)
        positions = dict(positions, **{node : coord + shift for node, coord in pos.items()})
        a = numpy.array(list(pos.values()))
        shift[1] -= numpy.max(abs(a[:,1])) + space
    return positions



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import sma
import networkx as nx

class MotifClassificator:
    """
    Abstract super class of all motif classificators. Roughly speaking, a classificator
    maps a motif to its motif class, e.g. (Peter, Lake, Forrest) to 'I.C'.
    
    This class cannot be used as classificator as it is abstract. Use one of its
    inheritors, e.g. :py:class:`sma.FourMotifClassificator`.
    
    Each instance of this class is callable mapping a motif to its class. For
    constructing such an object a SEN must be provided. The syntax is as follows,
    cf. :py:class:`sma.ThreeMotifClassificator`.
    
    .. code-block :: Python
    
        import sma    
        # Let G be some SEN
        motif = ('Peter', 'Lake', 'Forrest')
        classificator = sma.ThreeMotifClassificator(G)
        clss = classificator(motif) # e.g. 'I.C'
    
    Each inheritor of this class must provide the following attributes. Note, that
    the values listed here are dummies. See the inheriting classes.
    """
    def __init__(self, G : nx.Graph):
        self.G = G
    def __call__(self, motif):
        raise NotImplementedError("sma.MotifClassificator is an abstract class which should not be used as MotifClassificator")
    def info(self):
        """
        Returns the motif database entry for the motifs classified by this classificator,
        i.e. a subclass of :py:class:`sma.MotifInfo`.
        """
        raise NotImplementedError("sma.MotifClassificator is an abstract class which should not be used as MotifClassificator")

def classify4Motif(G : nx.Graph, motif) -> str:
    """
    Classifies a 4-motif given as a 4-tuple of its nodes.
    
    See :py:meth:`sma.binaryCodeToClass4Motifs`.
    
    :param G: the graph
    :param motif: the motif given as 4-tuple of its nodes
    :returns: class of the motif as string, e.g. 'I.C'
    """
    s1,s2,e1,e2 = motif
    binary = sma.binaryArrayHornersMethod([
            G.has_edge(s1, s2),
            G.has_edge(s2, e2),
            G.has_edge(e1, e2),
            G.has_edge(s1, e1),
            G.has_edge(s1, e2),
            G.has_edge(s2, e1)
        ])
    return sma.binaryCodeToClass4Motifs(binary)

class FourMotifClassificator(MotifClassificator):
    """
    :py:class:`sma.MotifClassificator` for classifying 4-motifs. Wrapps
    :py:meth:`sma.classify4Motif`.
    """
    def __call__(self, motif):
        return classify4Motif(self.G, motif)
    def info(self):
        return sma.Motif4Info

def classify3Motif(G : nx.Graph, motif) -> str:
    """
    Classifies a 3-motif given as a 3-tuple of its nodes.
    
    See :py:meth:`sma.binaryCodeToClass3Motifs`.
    
    :param G: the graph
    :param motif: the motif given as 3-tuple of its nodes
    :returns: class of the motif as string, e.g. 'I.C'
    """
    a, b1, b2 = motif
    binary = sma.binaryArrayHornersMethod([
            G.has_edge(b1,b2),
            G.has_edge(a, b1),
            G.has_edge(a, b2)
            ])
    return sma.binaryCodeToClass3Motifs(binary)

class ThreeMotifClassificator(MotifClassificator):
    """
    :py:class:`sma.MotifClassificator` for classifying 3-motifs such as 3E- or
    3S-motifs. Wrapps :py:meth:`sma.classify3Motif`. See also 
    :py:class:`sma.ThreePMotifClassificator` for plain motifs.
    """
    def __call__(self, motif):
        return classify3Motif(self.G, motif)
    def info(self):
        return sma.Motif3Info


def classify2Motif(G : nx.Graph, motif) -> bool:
    """
    Classifies 2-motifs. These motifs consist of two vertices. Two classes of
    motifs occur. Either the two vertices are linked (type 1) or the two vertices 
    not linked (type 0).
    
    :param G: the SEN
    :param motif: a pair of two vertices, the motif
    """
    return int(G.has_edge(*motif))

class TwoMotifClassificator(MotifClassificator):
    """
    :py:class:`sma.MotifClassificator` for classifying 2-motifs. See 
    :py:meth:`sma.classify2Motif`.
    """
    def __call__(self, motif):
        return classify2Motif(self.G, motif)
    def info(self):
        return sma.Motif2Info

    
def classify3pMotif(G : nx.Graph, motif) -> int:
    """
    Classifies a plain 3-motif given as a 3-tuple of its nodes.
    
    Since the three vertices in the motif are not distinguishable, there exist 
    four classes of motifs:
        
        - type 0: no edges
        - type 1: one edge
        - type 2: two edges
        - type 3: three edges
    
    :param G: the graph
    :param motif: the motif given as 3-tuple of its nodes
    :returns: class of the motif as integer
    """
    return (G.has_edge(motif[0], motif[1]) 
            + G.has_edge(motif[1], motif[2])
            + G.has_edge(motif[2], motif[0]))
    
class ThreePMotifClassificator(MotifClassificator):
    """
    :py:class:`sma.MotifClassificator` for classifying plain 3-motifs. These are
    motifs that do not respect whether the individual nodes are social or ecological.
    
    See :py:class:`sma.ThreeMotifClassificator` for the classificator that respects
    the types of the nodes.
    """
    def __call__(self, motif):
        return classify3pMotif(self.G, motif)
    def info(self):
        return sma.Motif3pInfo
    
class MultiMotifClassificator(MotifClassificator):
    """
    :py:class:`sma.MotifClassificator` for classifying multi-motifs. This class
    is a wrapper class for several other motif classificators. Based on the given
    arities, i.e. the numbers of nodes from the different levels of the SEN, a
    suitable classificator is choosen automatically. This means that depending on
    the given parameters, the number of classes and their names is different.
    
    See the documentation or query the motif database with :py:meth:`sma.motifInfo`
    and :py:meth:`sma.supportedSignatures` for a full list of supported motifs.
    
    Every instances of this class is aware of the names and amount of classes that
    it uses for classification.
    
    See also :py:meth:`sma.matchPositions` and :py:meth:`sma.multiSignature` for
    a description of the internal position matching mechanisms
    
    :param arities: arities
    :param roles: list of indices
    :raises NotSupportedError: whenever there is no matching classificator available
    """
    def __init__(self, G : nx.Graph, *arities : int, roles = []):
        super().__init__(G)
        self.arities = arities
        self.roles = roles
        signature_ordered = sma.multiSignature(arities)
        self.motif_info = sma.motifInfo(signature_ordered)
        self.signature = self.motif_info.signature
        self.positions = sma.matchPositions(self.signature, arities, roles)
        # initialize classificator
        self.classificator = self.motif_info.classificator(self.G)
    def __call__(self, motif):
        twisted_motif = []
        for p in self.positions:
            twisted_motif += motif[p]
        return self.classificator(twisted_motif)
    def __str__(self):
        return 'MultiMotifClassificator%s' % str(self.signature)
    def info(self):
        return self.motif_info

# Underneath you find very messy classificators for multi-motifs.

class Multi111MotifClassificator(MotifClassificator):
    def __call__(self, motif):
        return classify111Motif(self.G, motif)
    def info(self):
        return sma.Motif111Info

def classify111Motif(G : nx.Graph, motif) -> int:
    a, b, c = motif
    return 4 * G.has_edge(a,c) + 2 * G.has_edge(b, c) + G.has_edge(a, b)

class Multi121MotifClassificator(MotifClassificator):
    def __call__(self, motif):
        return classify121Motif(self.G, motif)
    def info(self):
        return sma.Motif121Info

def classify121Motif(G : nx.Graph, motif) -> int:
    a, b1, b2, c = motif
    if G.has_edge(b1, c) and G.has_edge(b2, c):
        if G.has_edge(a, c) and G.has_edge(b1, a) and G.has_edge(b2, a):
            # induced subgraphs {a, b1, c} and {a, b2, c} complete
            if G.has_edge(b1, b2):
                return 4 # Ohio E closed
            else:
                return 3 # Ohio E open
        if not(G.has_edge(b1, b2)) and not(G.has_edge(a,c)):
            t = G.has_edge(a, b1) + G.has_edge(a, b2)
            if t == 1:
                return 1 # (i)
            elif t == 2:
                return 2 # (h)
    return -1

class Multi221MotifClassificator(MotifClassificator):
    def __call__(self, motif):
        return classify221Motif(self.G, motif)
    def info(self):
        return sma.Motif221Info

def classify221Motif(G : nx.Graph, motif) -> str:
    a1, a2, b1, b2, c = motif
    if G.has_edge(a1, c) ^ G.has_edge(a2, c):
        return 'Unclassified'
    if G.has_edge(b1, c) ^ G.has_edge(b2, c):
        return 'Unclassified'
    block4 = classify4Motif(G, (a1, a2, b1, b2))
    if G.has_edge(a1, c):
        if G.has_edge(b1, c):
            return '%s.%d' % (block4, 3) # both bound
        else:
            return '%s.%d' % (block4, 1) # only (a) bound
    else:
        if G.has_edge(b1, c):
            return '%s.%d' % (block4, 2) # only (b) bound
        else:
            return '%s.%d' % (block4, 0) # nothing bound
    return 'Unclassified'

class Multi222MotifClassificator(MotifClassificator):
    def __call__(self, motif):
        return classify222Motif(self.G, motif)
    def info(self):
        return sma.Motif222Info    

def classify222Motif(G : nx.Graph, motif) -> int:
    a1, a2, b1, b2, c1, c2 = motif
    if G.has_edge(b1, b2) or not(G.has_edge(c1,c2)):
        return -1
    AC = (G.has_edge(a1, c1) + G.has_edge(a2, c2) +
          G.has_edge(a1, c2) + G.has_edge(a2, c1))
    if AC == 4: # all a-c edges are there
        if (G.has_edge(a1, b1) and G.has_edge(a2, b2) and
            G.has_edge(a1, b2) and G.has_edge(a2, b1) and
            G.has_edge(b1, c1) and G.has_edge(b2, c2) and
            G.has_edge(b1, c2) and G.has_edge(b2, c1)):
            if G.has_edge(a1, a2):
                return 4 # Ohio H closed
            else:
                return 3 # Ohio H open
        return -1
    elif AC == 0: # no a-c edges
        if not(G.has_edge(a1, b1)): # swap b1 and b2 
            h = b1
            b1 = b2
            b2 = h
        if not(G.has_edge(b1, c1)): # swap c1 and c2
            h = c1
            c1 = c2
            c2 = h
        if (G.has_edge(a1, b1) and G.has_edge(b1, c1) and
            G.has_edge(a2, b2) and G.has_edge(b2, c2) and
            not(G.has_edge(a1, b2)) and not(G.has_edge(a2, b1)) and
            not(G.has_edge(b1, c2)) and not(G.has_edge(b2, c1))):
            if G.has_edge(a1, a2):
                return 2 # (m)
            else:
                return 1 # (n)
    return -1


# -*- coding: utf-8 -*-

import numpy

NODE_TYPE_SOC = 1
NODE_TYPE_ECO = 0 

EDGE_TYPE_SOC_SOC = 2
EDGE_TYPE_ECO_SOC = 1
EDGE_TYPE_ECO_ECO = 0

MOTIF4_NAMES = ['I.A', 'I.B', 'I.C', 'I.D', 'II.A', 'II.B', 'II.C', 'II.D', 
                'III.A', 'III.B', 'III.C', 'III.D', 'IV.A', 'IV.B', 'IV.C', 'IV.D', 
                'V.A', 'V.B', 'V.C', 'V.D', 'VI.A', 'VI.B', 'VI.C', 'VI.D', 
                'VII.A', 'VII.B', 'VII.C', 'VII.D']
MOTIF3_NAMES = ['I.A', 'I.B', 'I.C', 'II.A', 'II.B', 'II.C']

COLORS_TYPES = {
                NODE_TYPE_ECO : 'green', 
                NODE_TYPE_SOC : 'red',
                2 : 'c', 
                3 : 'm', 
                4 : 'y', 
                5 : 'k', 
                6 : 'w'
                }

MULTI_DEFAULT_NAMES = ['Lake', 'Town', 'Issue', 'System', 'Entity', 'Item', 'Universe']

MOTIF3_EDGES = numpy.array([[0,0],[0,1],[0,2],[1,0],[1,1],[1,2]])

MOTIF4_EDGES = numpy.array([[0,2,0],
                         [1,2,0],
                         [0,2,1],
                         [1,2,1],
                         [0,2,0],
                         [1,2,0],
                         [0,2,1],
                         [1,2,1],
                         [0,4,0],
                         [1,4,0],
                         [0,4,1],
                         [1,4,1],
                         [0,0,0],
                         [1,0,0],
                         [0,0,1],
                         [1,0,1],
                         [0,3,0],
                         [1,3,0],
                         [0,3,1],
                         [1,3,1],
                         [1,1,0],
                         [1,1,1],
                         [1,2,0],
                         [1,2,1],
                         [0,1,0],
                         [0,1,1],
                         [0,2,0],
                         [0,2,1]])

MOTIF4_SYMMETRIES = numpy.array([2, 2, 2, 2, 
                                 2, 2, 2, 2, 
                                 4, 4, 4, 4, 
                                 1, 1, 1, 1, # sparse motifs IV.A-IV.D
                                 3, 3, 3, 3, 
                                 1, 1, 2, 2, 
                                 1, 1, 2, 2])

MOTIF4_AUT = numpy.array([2, 2, 2, 2, 
                          2, 2, 2, 2, 
                          1, 1, 1, 1, 
                          1, 1, 1, 1, # sparse motifs IV.A-IV.D
                          4, 4, 4, 4, 
                          4, 4, 2, 2, 
                          4, 4, 2, 2])
MOTIF3_AUT = numpy.array([1, 2, 1, 1, 2, 1])

# models for random baselines
MODEL_ERDOS_RENYI     = 'erdos_renyi'
MODEL_ACTORS_CHOICE   = 'actors_choice'
MODEL_FIXED_DENSITIES = 'fixed_densities'

def binaryCodeToClass4Motifs(digits : int) -> str:
    """
    Classifies 4-motifs according to 
    Ö. Bodin, M. Tengö: Disentangling intangible social–ecological systems 
    Global Environmental Change 22 (2012) 430–439
    http://dx.doi.org/10.1016/j.gloenvcha.2012.01.005
    
    :param digits: binary code of the motif as described in the paper
    :returns: string code 'I.A' to 'VII.D'
    """
    if digits == 0b000000:  return 'IV.A'
    if(digits == 0b000100 or
       digits == 0b010000 or
       digits == 0b000001 or
       digits == 0b000010): return 'VII.A'
    if digits == 0b001000:  return 'IV.C'
    if digits == 0b100000:  return 'IV.B'
    if(digits == 0b000101 or
       digits == 0b010010): return 'II.A'
    if(digits == 0b000110 or
       digits == 0b010001): return 'VII.C'
    if(digits == 0b010100 or
       digits == 0b000011): return 'I.A'
    if(digits == 0b001100 or
       digits == 0b011000 or
       digits == 0b001010 or
       digits == 0b001001): return 'VII.B'
    if(digits == 0b100100 or
       digits == 0b100001 or
       digits == 0b100010 or
       digits == 0b110000): return 'VI.A'
    if digits == 0b101000:  return 'IV.D'
    if(digits == 0b001101 or
       digits == 0b011010): return 'II.C'
    if(digits == 0b100101 or
       digits == 0b110010): return 'II.B'
    if(digits == 0b101100 or
       digits == 0b111000 or
       digits == 0b101001 or
       digits == 0b101010): return 'VI.B'
    if(digits == 0b001110 or
       digits == 0b011001): return 'VII.D'
    if(digits == 0b010110 or
       digits == 0b010011 or
       digits == 0b010101 or
       digits == 0b000111): return 'V.A'
    if(digits == 0b100110 or
       digits == 0b110001): return 'VI.C'
    if(digits == 0b011100 or
       digits == 0b001011): return 'I.C'
    if(digits == 0b110100 or
       digits == 0b100011): return 'I.B'
    if digits == 0b010111:  return 'III.A'
    if(digits == 0b101101 or
       digits == 0b111010): return 'II.D'
    if(digits == 0b011110 or
       digits == 0b011011 or
       digits == 0b011101 or
       digits == 0b001111): return 'V.C'
    if(digits == 0b101110 or
       digits == 0b111001): return 'VI.D'
    if(digits == 0b110110 or
       digits == 0b110011 or
       digits == 0b110101 or
       digits == 0b100111): return 'V.B'
    if(digits == 0b111100 or
       digits == 0b101011): return 'I.D'
    if digits == 0b011111:  return 'III.C'
    if digits == 0b110111:  return 'III.B'
    if(digits == 0b111110 or
       digits == 0b111101 or
       digits == 0b101111 or
       digits == 0b111011): return 'V.D'
    if digits == 0b111111:  return 'III.D'
    
    assert digits >= 0 and digits < 64, 'motif code out of range {0:b}'.format(digits)
    
    return None

def binaryCodeToClass3Motifs(digits : int) -> str:
    """
    Classifies 3-motifs according to the following scheme: A 3-motif consists of either
    one social and two ecological nodes or one ecological and two social nodes. Thus, 
    one node can be considered distinct from the two other nodes. Let's call this node
    a and the other two nodes b1 and b2. The following classes occur (listing undirected edges):
        
    class I: without edge (b1,b2):
        - I.A: no edges,
        - I.B: either (a,b1) or (a,b2),
        - I.C: both (a,b1) and (a,b2).
    class II: with edge (b1,b2):
        - II.A: (b1,b2),
        - II.B: (b1,b2) and either (a,b1) or (a,b2),
        - II.C: (b1,b2) and both (a,b1) and (a,b2).
    
    :param digits: binary code of the motif as described above
    :returns: string code 'I.A' to 'VII.D'
    """ 
    assert digits >= 0 and digits < 8, 'motif code out of rande {0:b}'.format(digits)
    cl1 = 'I' if digits // 4 == 0 else 'II'
    cl2 = ''
    digits2 = digits % 4
    if digits2 == 0b00:
        cl2 = 'A'
    elif digits2 == 0b11:
        cl2 = 'C'
    else:
        cl2 = 'B'
    return cl1 + '.' + cl2


#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import sma
import itertools
import networkx as nx
import multiprocessing

class _simulateBaselineAutoMapper:
    def __init__(self, motifs, assume_sparse):
        self.motifs = motifs
        self.assume_sparse = assume_sparse
    def __call__(self, G):
        partial, _ = sma.countMotifsAuto(G, *self.motifs, assume_sparse = self.assume_sparse)
        return partial

def simulateBaselineAuto(G : nx.Graph, 
                         *motifs, 
                         n : int= 100,
                         processes : int = 0,
                         chunksize : int = 100,
                         assume_sparse : bool = False,
                         model : str = sma.MODEL_ERDOS_RENYI) -> dict:
    """
    Simulates a random baseline of graphs similar to a given SEN and counts motifs
    in these randomly generated graphs.
    
    See :py:meth:`sma.randomSimilarMultiSENs` for a documentation of the various
    baseline models.
    
    :param G: the SEN
    :param motifs: motif identifier strings of motifs that shall be counted
    :param n: number of iterations
    :param processes: number of processes, default zero (no multiprocessing)
    :param chunksize: chunksize for multiprocessing
    :param assume_sparse: whether the random graphs shall be assumed to be sparse.
        Used to find an ideal counter, cf. :py:meth:`sma.findIdealCounter`.
    :param model: model to be used, default :py:const:`sma.MODEL_ERDOS_RENYI`.
    :returns: dict mapping motif identifiers to list of counts
    """
    random_networks = sma.randomSimilarMultiSENs(G, model = model)
    source = itertools.islice(random_networks, n)
    mapper = _simulateBaselineAutoMapper(motifs, assume_sparse)
    result = {k : [] for k in motifs}
    if processes <= 0:
        for row in map(mapper, source):
            for k, v in row.items():
                result[k].append(v)
            del row
    else:
        with multiprocessing.Pool(processes) as p:
            rows = p.imap_unordered(mapper, source, chunksize = chunksize)
            for row in rows:
                for k, v in row.items():
                    result[k].append(v)
                del row
            p.close()
            p.join()        
    return result
                
            
    


#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import sma
import networkx as nx
from scipy.special import binom
import numpy
import functools, itertools

def _probability(edges, sym, maxedges, probs):
    prob = sym
    for e, p, m in zip(edges, probs, maxedges):
        prob *= p**e * (1-p)**(m-e)
    return prob

def expectedMultiMotifs(G : nx.Graph,
                        *arities, 
                        roles = [],
                        **kwargs):
    """
    Front-end function for several functions for computing motif expectations. It supports
    the same parameters for defining the motifs (arities, roles) as other multi-level
    (counting) functions. Check the documentations of the respective functions.
    The supported functions are:
        
        - :py:meth:`sma.expected3Motifs`
        - :py:meth:`sma.expected4Motifs`
        - :py:meth:`sma.expected121Motifs`
        - :py:meth:`sma.expected221Motifs`
        - :py:meth:`sma.expected222Motifs`
    
    See the documentation for an exposition of the mathematical background.
    
    :param G: the SEN
    :param arities: arities of the motifs, cf. :py:class:`sma.MultiMotifClassificator`.
    :param roles: roles of the levels, cf. :py:class:`sma.MultiMotifClassificator`.
    :param kwargs: additional parameters for the functions, see above.
    """
    signature  = sma.multiSignature(arities)
    motif_info = sma.motifInfo(signature)
    positions  = sma.matchPositions(motif_info.signature, arities, roles)
    if motif_info.expectations is None:
        raise NotImplementedError("expectations for %s-motifs not implemented" % str(motif_info.signature))
    return motif_info.expectations(G, *positions, **kwargs)
    
def varMultiMotifs(G : nx.Graph,
                        *arities, 
                        roles = [],
                        **kwargs):
    """
    Front-end function for several functions for computing motif variances. It supports
    the same parameters for defining the motifs (arities, roles) as other multi-level
    (counting) functions. Check the documentations of the respective functions.
    The supported functions are:
        
        - :py:meth:`sma.var3Motifs`
    
    See the documentation for an exposition of the mathematical background.
    
    :param G: the SEN
    :param arities: arities of the motifs, cf. :py:class:`sma.MultiMotifClassificator`.
    :param roles: roles of the levels, cf. :py:class:`sma.MultiMotifClassificator`.
    :param kwargs: additional parameters for the functions, see above.
    """
    signature  = sma.multiSignature(arities)
    motif_info = sma.motifInfo(signature)
    positions  = sma.matchPositions(motif_info.signature, arities, roles)
    if motif_info.variances is None:
        raise NotImplementedError("variances for %s-motifs not implemented" % str(motif_info.signature))
    return motif_info.variances(G, *positions, **kwargs)

def distributionMotifsAuto(G : nx.Graph, 
                           *motifs : str, 
                           model : str = sma.MODEL_ERDOS_RENYI,
                           level : int = -1):
    """
    Front-end function for functions in sma which provide insight into the distribution
    of motif counts in various baseline models.
    
    This function returns a tuple consisting of two dicts:
        
        1. a dict of partial results as requested using the motif identifiers. This
           dict maps motif identifiers to tuples containing firstly the expectation
           and secondly the variance of this motif.
        2. a dict of the results of all motifs with same arities/roles as the motif
           specified using the motif identifiers. This dict maps the heads of the
           motif identifers to dicts mapping motif classes to tuples of expectations
           and variances as above.
           
    Values which are not available in the selected model are given as ``None``.
    
    The following models (specify using parameter ``model``) are supported:
        
        - ``erdos_renyi`` (default) general Erdős-Rényi random graphs with fixed
          edge probabilities for each subsystem equalling the densities of the 
          given SEN, see documentation, cf. :py:const:`sma.MODEL_ERDOS_RENYI`.
        - ``actors_choice`` edges in one specific level are chosen at random. The
          remainder of the network is fixed. The edge probability in the variable
          level is extracted from the given SEN. Use parameter ``level`` to specify
          the ``sesType`` of the variable level. If :math:`-1` (default), then
          the first entry equalling :math:`2` in the list of arities is selected,
          cd. :py:const:`sma.MODEL_ACTORS_CHOICE`.
    
    :param G: the SEN
    :param motifs: motif identifier strings
    :param model: model to be used
    :param level: additional parameter for model ``actors_choice``
    :returns: tuple of two dicts as above
    :raises TypeError: when the motif identifiers cannot be parsed
    """
    grouped_request = sma.groupMotifIdentifiers(*motifs)
    partial_results = {}
    total_results = {}
    for head in grouped_request.keys():
        arities, roles, _ = sma.parseMotifIdentifier(head)
        if model == sma.MODEL_ERDOS_RENYI:
            expectations = expectedMultiMotifs(G, *arities, roles = roles)
            variances = None
            try:
                variances = varMultiMotifs(G, *arities, roles = roles)
            except NotImplementedError:
                # variance is not implemented for most motifs, so better pass
                pass
            total_results[head] = {k : (expectations[k] if expectations is not None else None, variances[k] if variances is not None else None) for k in expectations.keys()}
        elif model == sma.MODEL_ACTORS_CHOICE:
            total_results[head]  = distributionMotifsActorsChoice(G,
                                                                    *arities,
                                                                    roles = roles,
                                                                    level = level,
                                                                    array = False)
        else:
            raise NotImplementedError("model %s not supported" % model)
        for motif in grouped_request[head]:
            key = motif
            if key not in total_results[head]:
                try:
                    if int(key) in total_results[head]:
                        key = int(key)
                    else:
                        raise TypeError("Unrecognizable motif identifier %s[%s]." % (head, motif))
                except ValueError:
                    pass
            try:
                partial_results["%s[%s]" % (head, motif)] = total_results[head][key]
            except KeyError:
                raise TypeError("Unrecognizable motif identifier %s[%s]." % (head, motif))
    return (partial_results, total_results)
    
def expected3Motifs(G : nx.Graph, 
                    level0, 
                    level1, 
                    array = False, 
                    densities : bool = False):
    """
    Computes the expected number of 3-motifs in a SEN. See the documentation for an
    exposition of the mathematical background.
    
    Use functions :py:meth:`sma.expected3EMotifs` and :py:meth:`sma.expected3SMotifs`
    for regular two level graphs.
    
    If ``densities = True``, the returned values sum up to one. Otherwise their
    sum is total number of 3-motifs, see :py:meth:`sma.total3EMotifs` etc.
    
    :param G: the SEN
    :param level0: ``sesType`` of the distinct node, e.g. in 3E-motifs :py:const:`sma.NODE_TYPE_ECO`
    :param level1: ``sesType`` of the other nodes, e.g. in 3E-motifs :py:const:`sma.NODE_TYPE_SOC`
    :param array: whether the result shall be an array or a dict
    :param densities: whether densities or absolute numbers of motifs shall be returned
    """
    v = sma.nodesCount(G)
    m = sma.maxEdgeCountMatrix(list(v.values()))
    d = sma.edgesCountMatrix(G)
    
    total = 1 if densities else binom(v[level1], 2) * v[level0]
    p0 = d[level0, level1] / m[level0, level1] if level0 <= level1 else d[level1,level0] / m[level1,level0]
    p1 = d[level1, level1] / m[level1, level1]
    
    maxedges = numpy.max(sma.MOTIF3_EDGES, axis = 0)
    res = list(itertools.starmap(functools.partial(_probability, maxedges = maxedges, probs = [p1, p0]),
                  zip(sma.MOTIF3_EDGES, sma.MOTIF3_AUT)))
    results = total * numpy.array(res)
    if not array:
        return {k : v for k,v in zip(sma.MOTIF3_NAMES, results)}
    return results

def expected4Motifs(G : nx.Graph, 
                    level0 = sma.NODE_TYPE_SOC, 
                    level1 = sma.NODE_TYPE_ECO, 
                    array = False, 
                    densities = False):
    """
    Computes the expected number of 4-motifs in a SEN. See the documentation for an
    exposition of the mathematical background.
    
    If ``densities = True``, the returned values sum up to one. Otherwise their
    sum is total number of 4-motifs, see :py:meth:`sma.total4Motifs` etc.
    
    :param G: the SEN
    :param level0: ``sesType`` of the upper nodes, standard :py:const:`sma.NODE_TYPE_SOC`
    :param level1: ``sesType`` of the lower nodes, e.g. in 3E-motifs :py:const:`sma.NODE_TYPE_ECO`
    :param array: whether the result shall be an array or a dict
    :param densities: whether densities or absolute numbers of motifs shall be returned
    """
    v = sma.nodesCount(G)
    m = sma.maxEdgeCountMatrix(list(v.values()))
    d = sma.edgesCountMatrix(G)
    
    total = 1 if densities else binom(v[level1], 2) * binom(v[level0], 2)
    p0 = d[level0, level0] / m[level0, level0]
    p1 = d[level0, level1] / m[level0, level1] if level0 <= level1 else d[level1,level0] / m[level1,level0]
    p2 = d[level1, level1] / m[level1, level1]
    maxedges = numpy.max(sma.MOTIF4_EDGES, axis = 0)

    res = list(itertools.starmap(functools.partial(_probability, maxedges = maxedges, probs = [p0, p1, p2]),
                  zip(sma.MOTIF4_EDGES, sma.MOTIF4_AUT)))
    results = total * numpy.array(res)
    if not array:
        return {k : v for k,v in zip(sma.MOTIF4_NAMES, results)}
    return results

def expected3EMotifs(G : nx.Graph, **kwargs):
    """
    Computes the expected number of 3E-motifs in a SEN. See the documentation for an
    exposition of the mathematical background. See also :py:meth:`sma.expected3Motifs`.
    
    :param G: the SEN
    :param kwargs: additional parameters for :py:meth:`sma.expected3Motifs`.
    """
    return expected3Motifs(G, sma.NODE_TYPE_ECO, sma.NODE_TYPE_SOC, **kwargs)
def expected3SMotifs(G : nx.Graph, **kwargs):
    """
    Computes the expected number of 3S-motifs in a SEN. See the documentation for an
    exposition of the mathematical background. See also :py:meth:`sma.expected3Motifs`.
    
    :param G: the SEN
    :param kwargs: additional parameters for :py:meth:`sma.expected3Motifs`.
    """
    return expected3Motifs(G, sma.NODE_TYPE_SOC, sma.NODE_TYPE_ECO, **kwargs)

def expected121Motifs(G : nx.Graph, 
                      level0, 
                      level1, 
                      level2, 
                      array = False,
                      density = False):
    """
    Computes the expected number of multi level motifs with arities 1, 2, 1. 
    See the documentation for an exposition of the mathematical background.
    
    :param G: the SEN
    :param level0: ``sesType`` of the upper node
    :param level1: ``sesType`` of the middle nodes
    :param level2: ``sesType`` of the lower node
    :param array: whether the result shall be an array or a dict
    :param densities: whether densities or absolute numbers of motifs shall be returned
    """
    v = sma.nodesCount(G)
    m = sma.maxEdgeCountMatrix(list(v.values()))
    d = sma.edgesCountMatrix(G)
    total = 1 if density else v[level0] * binom(v[level1], 2) * v[level2]
    p01 = d[level0, level1] / m[level0, level1] if level0 <= level1 else d[level1,level0] / m[level1,level0]
    p02 = d[level0, level2] / m[level0, level2] if level0 <= level2 else d[level2,level0] / m[level2,level0]
    p11 = d[level1, level1] / m[level1, level1]
    p12 = d[level1, level2] / m[level1, level2] if level1 <= level2 else d[level2,level1] / m[level2,level1]
    results = numpy.array([
            0,
            2 * p01 * (1-p01) * (1-p11) * p12**2 * (1-p02),
            p01**2 * (1-p11) * p12**2 * (1-p02),
            p01**2 * (1-p11) * p12**2 * p02,
            p01**2 * p11 * p12**2 * p02,
        ])
    results[0] = 1 - sum(results)
    results = total * results
    if array:
        return results
    return {k : r for k, r in zip([-1,1,2,3,4], results)}

def expected221Motifs(G : nx.Graph, 
                      level0, 
                      level1, 
                      level2, 
                      array = False,
                      density = False):
    """
    Computes the expected number of multi level motifs with arities 2, 2, 1. 
    See the documentation for an exposition of the mathematical background.
    
    Uses :py:meth:`sma.expected4Motifs` for computing the expectations of the upper
    part.
    
    :param G: the SEN
    :param level0: ``sesType`` of the upper node
    :param level1: ``sesType`` of the middle nodes
    :param level2: ``sesType`` of the lower node
    :param array: whether the result shall be an array or a dict
    :param densities: whether densities or absolute numbers of motifs shall be returned
    """
    v = sma.nodesCount(G)
    m = sma.maxEdgeCountMatrix(list(v.values()))
    d = sma.edgesCountMatrix(G)
    total = 1 if density else binom(v[level0], 2) * binom(v[level1], 2) * v[level2]
    p02 = d[level0, level2] / m[level0, level2] if level0 <= level2 else d[level2,level0] / m[level2,level0]
    p12 = d[level1, level2] / m[level1, level2] if level1 <= level2 else d[level2,level1] / m[level2,level1]
    
    values4 = sma.expected4Motifs(G, level0, level1, array = True, densities = True)
    
    factors = [
            (1-p02)**2 * (1-p12)**2,
            p02**2 * (1-p12)**2,
            (1-p02)**2 * p12**2,
            p02**2 * p12**2,
        ]
    result = {'Unclassified' : 0, **{
            '%s.%d' % (m, i) : total * f * v
            for (m, v), (i, f) in itertools.product(zip(sma.MOTIF4_NAMES, values4), zip([0, 1, 2, 3], factors))
        }}
    result['Unclassified'] = total - sum(result.values())
    if array:
        return numpy.array(list(result.values()))
    return result

def expected222Motifs(G : nx.Graph, 
                      level0, 
                      level1, 
                      level2, 
                      array = False,
                      density = False):
    """
    Computes the expected number of multi level motifs with arities 2, 2, 2. 
    See the documentation for an exposition of the mathematical background.
    
    :param G: the SEN
    :param level0: ``sesType`` of the upper node
    :param level1: ``sesType`` of the middle nodes
    :param level2: ``sesType`` of the lower node
    :param array: whether the result shall be an array or a dict
    :param densities: whether densities or absolute numbers of motifs shall be returned
    """
    v = sma.nodesCount(G)
    m = sma.maxEdgeCountMatrix(list(v.values()))
    d = sma.edgesCountMatrix(G)
    total = 1 if density else binom(v[level0], 2) * binom(v[level1], 2) * binom(v[level2], 2)
    p00 = d[level0, level0] / m[level0, level0]
    p01 = d[level0, level1] / m[level0, level1] if level0 <= level1 else d[level1,level0] / m[level1,level0]
    p02 = d[level0, level2] / m[level0, level2] if level0 <= level2 else d[level2,level0] / m[level2,level0]
    p11 = d[level1, level1] / m[level1, level1]
    p12 = d[level1, level2] / m[level1, level2] if level1 <= level2 else d[level2,level1] / m[level2,level1]
    p22 = d[level2, level2] / m[level2, level2]

    results = numpy.array([
            0,
            2 * (1-p00) * (1-p01)**2 * p01**2 * (1-p12)**2 * (1-p11) * p12**2 * p22 * (1-p02)**4,
            2 * p00 * (1-p01)**2 * p01**2 * (1-p12)**2 * (1-p11) * p12**2 * p22 * (1-p02)**4,
            (1-p00) * p01**4 * p02**4 * (1-p11) * p12**4 * p22,
            p00 * p01**4 * p02**4 * (1-p11) * p12**4 * p22
        ])
    results[0] = 1 - sum(results)
    results = total * results
    if array:
        return results
    return {k : r for k, r in zip([-1,1,2,3,4], results)}

def var3Motifs(G : nx.Graph, 
               level0, 
               level1, 
               array : bool = False,
               second_moment : bool = False):
    """
    Computes variances of the number for 3-motifs in a SEN. See documentation for
    the mathematical background.
    
    This function uses the following formula where :math:`X` denotes the random
    variable number of 3-motifs
    
    .. math::
        
        \mathbb{V}[X] = \mathbb{E}[X^2] - (\mathbb{E}[X])^2
        
    
    and computes the expectation using :py:meth:`sma.expected3Motifs`. Set
    ``second_moment = True`` to compute :math:`\mathbb{E}[X^2]` instead of the variance.
    In this case :py:meth:`sma.expected3Motifs` is not called.
    
    :param G: the SEN
    :param level0: ``sesType`` of the distinct node, e.g. in 3E-motifs :py:const:`sma.NODE_TYPE_ECO`
    :param level1: ``sesType`` of the other nodes, e.g. in 3E-motifs :py:const:`sma.NODE_TYPE_SOC`
    :param array: whether the result should be an array instead of a dict
    :param second_moment: switch on for computing the second moment instead of the
        variance
    """
    m = sma.maxEdgeCountMatrix(sma.nodesCount(G, array = True))
    d = sma.edgesCountMatrix(G)
    v = sma.nodesCount(G)
    p0 = d[level0, level1] / m[level0, level1] if level0 <= level1 else d[level1,level0] / m[level1,level0]
    p1 = d[level1, level1] / m[level1, level1]
    v0 = v[level0]
    v1 = v[level1]     
    partition = numpy.array([
                 v0 * binom(v1,2),                           # all nodes shared
                 v0 * v1 * (v1-1) * (v1-2),                  # one from each level shared
                 binom(v1, 2) * v0 * (v0-1),                 # two from non-distinct level shared
                 v0 * v1 * (v0-1) * (v1-1) * (v1-2),         # one non-distinct node shared
                 v0 * binom(v1,2) * binom(v1-2,2),           # distinct node shared
                 v0 * (v0-1) * binom(v1,2) * binom(v1-2,2)   # no nodes shared
            ])
    # one column for each partion, see above
    # one row for each motif I.A-II.C
    motifprops = numpy.array([
                [(1-p0)**2 * (1-p1), (1-p0)**3*(1-p1)**2,                   (1-p0)**4 * (1-p1),        (1-p0)**4 * (1-p1)**2,       (1-p0)**4 * (1-p1)**2,       (1-p0)**4 * (1-p1)**2],
                [2*(1-p0)*(1-p1)*p0, (1-p1)**2*((1-p0)**2*p0+p0**2*(1-p0)), 4*(1-p0)**2*p0**2 *(1-p1), 4*(1-p0)**2*p0**2*(1-p1)**2, 4*(1-p0)**2*p0**2*(1-p1)**2, 4*(1-p0)**2*p0**2*(1-p1)**2],
                [p0**2*(1-p1),       p0**3*(1-p1)**2,                       p0**4*(1-p1),              p0**4*(1-p1)**2,             p0**4*(1-p1)**2,             p0**4*(1-p1)**2],
                [(1-p0)**2 *p1,      (1-p0)**3*p1**2,                       (1-p0)**4*p1,              (1-p0)**4 * p1**2,           (1-p0)**4 * p1**2,           (1-p0)**4 * p1**2],
                [2*(1-p0)*p1*p0,     p1**2*((1-p0)**2*p0+p0**2*(1-p0)),     4*(1-p0)**2*p0**2 * p1,    4*(1-p0)**2*p0**2*p1**2,     4*(1-p0)**2*p0**2*p1**2,     4*(1-p0)**2*p0**2*p1**2],
                [p0**2*p1,           p0**3*p1**2,                           p0**4*p1,                  p0**4*p1**2,                 p0**4*p1**2,                 p0**4*p1**2]
            ])
    if second_moment:
        result = motifprops @ partition
    else:
        expectations = expected3Motifs(G, level0, level1, array=True)
        result = motifprops @ partition - expectations**2
    if array:
        return result
    else:
        return {k : v for k, v in zip(sma.MOTIF3_NAMES, result)}
    
def var3EMotifs(G : nx.Graph, **kwargs):
    """
    Computes the variance of the number of 3E-motifs in a SEN. See the documentation for a
    exposition of the mathematical background. See also :py:meth:`sma.expected3Motifs`.
    
    :param G: the SEN
    :param kwargs: additional parameters for :py:meth:`sma.var3Motifs`.
    """
    return var3Motifs(G, sma.NODE_TYPE_ECO, sma.NODE_TYPE_SOC, **kwargs)

def var3SMotifs(G : nx.Graph, **kwargs):
    """
    Computes the variance of the number of 3S-motifs in a SEN. See the documentation for a
    exposition of the mathematical background. See also :py:meth:`sma.expected3Motifs`.
    
    :param G: the SEN
    :param kwargs: additional parameters for :py:meth:`sma.var3Motifs`.
    """
    return var3Motifs(G, sma.NODE_TYPE_SOC, sma.NODE_TYPE_ECO, **kwargs)

def distributionMotifsActorsChoice(G : nx.Graph, 
                                   *arities, 
                                   roles = [], 
                                   level : int = -1,
                                   array : bool = False):
    """
    Computes expectation and variances for Actor's Choice distribution of motifs.
    See the documentation for a description of this model.
    
    If ``level = -1``, then the first entry equalling two in the signature of
    the motif is taken as level.
    
    :param G: the SEN
    :param arities: arities of the (multi-level) motif
    :param level: level which is assumed to be variable in the Actor's Choice Model
    :param array: whether the output is a tuple ``(e,v)`` where ``e`` is a vector
        of expectations and ``v`` a vector of variances, or dict mapping motif names
        to tuples of expectation and variance for this motif
    :returns: properties of the distribution as tuple or dict (see ``array``)
    """
    signature  = sma.multiSignature(arities)
    motif_info = sma.motifInfo(signature)
    positions  = sma.matchPositions(motif_info.signature, arities, roles)
    
    if level < 0:
        assert 2 in signature, 'signature does not contain two'
        level = positions[signature.index(2)]
    
    contributions, _    = sma.edgeContributionList(G, *motif_info.signature, roles = positions, level = level)
    p                   = sma.densityMatrix(G)[level,level]
    sums                = numpy.sum(contributions, axis = 1)
    squares             = numpy.sum(contributions**2, axis = 1)
    expectations_open   = (1-p) * sums
    expectations_closed = p * sums
    variances           = p * (1-p) * squares
    
    # de-group open and closed motifs
    indexer = {motif : i for i, motif in enumerate(motif_info.classes)}
    abstract_level = positions.index(level)
    if abstract_level not in motif_info.projections:
        raise NotImplementedError('ActorsChoice not implemented for signature %s on level %d' \
                                  % (str(motif_info.signature), level))
    grouper = motif_info.projections[abstract_level]
    e = numpy.zeros((len(indexer),))
    v = numpy.zeros((len(indexer),))
    for (op, cl), eo, ec, var in zip(grouper, expectations_open, expectations_closed, variances):
        e[indexer[op]] = eo
        e[indexer[cl]] = ec
        v[indexer[op]] = v[indexer[cl]] = var
    
    if array:
        return (e,v)
    else:
        return {k : (exp, var) for (k, exp, var) in zip(motif_info.classes, e, v)}


# Statistical helper function
def markovBound(value : numpy.ndarray, expectation : numpy.ndarray) -> numpy.ndarray:
    """
    Uses Markov's inequality to compute bounds for the probability that a random
    variable with given expectation takes values higher than the given values.
    
    .. math::
        
        \mathbb{P}[X \geq a] \leq \\frac{\mathbb{E}[X]}{a} 
        
    where :math:`a` is given by ``values`` and :math:`\mathbb{E}[X]` is given
    by ``expectation``.
    
    See `Markov's inequality on Wikipedia <https://en.wikipedia.org/wiki/Markov%27s_inequality>`_.
    
    :param value: the value
    :param expectation: the expectation
    :returns: bounds on the probabilities
    """
    return expectation / value

def cantelliBound(value : numpy.ndarray, 
                  expectation : numpy.ndarray, 
                  variance : numpy.ndarray) -> numpy.ndarray:
    """
    Uses Cantelli's inequality to compute bounds for the probability that a random
    variable with given expectation takes values higher/lower than the given values.
    
    .. math::
        :nowrap:
        
        \[
        \mathbb{P}[X - \mathbb{E}[X]\geq a] \leq 
        \\begin{cases}
        \\frac{\sigma^2}{\sigma^2 + a^2}, & a > 0 \\\\
        1 - \\frac{\sigma^2}{\sigma^2 + a^2}, & a < 0 \\\\ \end{cases}.
        \]
    
    where :math:`a` is given by ``value`` minus ``expectation`` and 
    :math:`\sigma^2` is given by ``variance``.
    
    See `Cantelli's inequality on Wikipedia <https://en.wikipedia.org/wiki/Cantelli%27s_inequality>`_.
    
    :param value: the value
    :param expectation: the expectation
    :returns: bounds on the probabilities
    """
    l = value - expectation
    bound = variance / (variance + l**2)
    signs = l < 0
    return signs * 1 + ((signs * -2)+1) * bound